#  Copyright (c) 2024 zfit
import jacobi
import numpy as np
import pytest
import tensorflow as tf

import zfit.core.basepdf
import zfit.models.dist_tfp
import zfit.settings
from zfit import z
import zfit.z.numpy as znp
from zfit.core.loss import UnbinnedNLL
from zfit.core.space import Space
from zfit.minimize import Minuit
from zfit.pdf import Gauss
from zfit.util.exception import BehaviorUnderDiscussion, IntentionAmbiguousError

mu_true = 1.2
sigma_true = 4.1
mu_true2 = 1.01
sigma_true2 = 3.5

yield_true = 3000
test_values_np = np.random.normal(loc=mu_true, scale=sigma_true, size=(yield_true, 1))


def create_test_values(size):
    return z.random.get_prng().normal(mean=mu_true, stddev=sigma_true, shape=(size, 1))


test_values_np2 = np.random.normal(loc=mu_true2, scale=sigma_true2, size=yield_true)

low, high = -24.3, 28.6


def create_params1(nameadd=""):
    mu1 = zfit.Parameter(
        "mu1" + nameadd, z.to_real(mu_true) - 0.2, mu_true - 1.0, mu_true + 1.0
    )
    sigma1 = zfit.Parameter(
        "sigma1" + nameadd,
        z.to_real(sigma_true) - 0.3,
        sigma_true - 2.0,
        sigma_true + 2.0,
    )
    return mu1, sigma1


def create_params2(nameadd=""):
    mu2 = zfit.Parameter(
        "mu25" + nameadd, z.to_real(mu_true) - 0.2, mu_true - 1.0, mu_true + 1.0
    )
    sigma2 = zfit.Parameter(
        "sigma25" + nameadd,
        z.to_real(sigma_true) - 0.3,
        sigma_true - 2.0,
        sigma_true + 2.0,
    )
    return mu2, sigma2


def create_params3(nameadd=""):
    mu3 = zfit.Parameter(
        "mu35" + nameadd, z.to_real(mu_true) - 0.2, mu_true - 1.0, mu_true + 1.0
    )
    sigma3 = zfit.Parameter(
        "sigma35" + nameadd,
        z.to_real(sigma_true) - 0.3,
        sigma_true - 2.0,
        sigma_true + 2.0,
    )
    yield3 = zfit.Parameter("yield35" + nameadd, yield_true + 300, 0, 10000000)
    return mu3, sigma3, yield3


obs1 = zfit.Space(
    "obs1",
    (
        np.min([test_values_np[:, 0], test_values_np2]) - 1.4,
        np.max([test_values_np[:, 0], test_values_np2]) + 2.4,
    ),
)

mu_constr = [1.6, 0.02]  # mu, sigma
sigma_constr = [3.5, 0.01]
constr = lambda: [mu_constr[1], sigma_constr[1]]
constr_tf = lambda: z.convert_to_tensor(constr())
covariance = lambda: np.array([[mu_constr[1] ** 2, 0], [0, sigma_constr[1] ** 2]])
covariance_tf = lambda: z.convert_to_tensor(covariance())


def create_gauss1(obs=obs1):
    mu, sigma = create_params1()
    return Gauss(mu, sigma, obs=obs, name="gaussian1"), mu, sigma


def create_gauss2(obs=obs1):
    mu, sigma = create_params2()
    return Gauss(mu, sigma, obs=obs, name="gaussian2"), mu, sigma


def create_gauss3ext():
    mu, sigma, yield3 = create_params3()
    gaussian3 = Gauss(mu, sigma, obs=obs1, name="gaussian3")
    gaussian3 = gaussian3.create_extended(yield3)
    return gaussian3, mu, sigma, yield3


def create_simpel_loss():
    _, mu1, sigma1 = create_gauss1()
    return zfit.loss.SimpleLoss(
        lambda x: (x[0] - 0.37) ** 4 * (x[1] - 2.34) ** 4,
        params=[mu1, sigma1],
        errordef=0.5,
    )


def create_simultaneous_loss():
    test_values = tf.constant(test_values_np)
    test_values = zfit.Data.from_tensor(obs=obs1, tensor=test_values)
    test_values2 = tf.constant(test_values_np2)
    test_values2 = zfit.Data.from_tensor(obs=obs1, tensor=test_values2)
    gaussian1, mu1, sigma1 = create_gauss1()
    gaussian2, mu2, sigma2 = create_gauss2()
    gaussian2 = gaussian2.create_extended(zfit.Parameter("yield_gauss2", 5))
    nll = zfit.loss.UnbinnedNLL(
        model=[gaussian1, gaussian2],
        data=[test_values, test_values2],
    )
    return mu1, mu2, nll, sigma1, sigma2


@pytest.mark.parametrize("size", [None, 5000, 50000, 300000, 3_000_000])
@pytest.mark.flaky(3)  # minimization can fail
def test_extended_unbinned_nll(size):
    if size is None:
        test_values = z.constant(test_values_np)
        size = test_values.shape[0]
    else:
        test_values = create_test_values(size)
    test_values = zfit.Data.from_tensor(obs=obs1, tensor=test_values)
    gaussian3, mu3, sigma3, yield3 = create_gauss3ext()
    nll = zfit.loss.ExtendedUnbinnedNLL(
        model=gaussian3, data=test_values, fit_range=(-20, 20)
    )
    assert {mu3, sigma3, yield3} == nll.get_params()
    minimizer = Minuit(tol=1e-4)
    status = minimizer.minimize(loss=nll)
    params = status.params
    assert pytest.approx(
        np.mean(test_values.value()), rel=0.05
    ) == params[mu3]["value"]
    assert pytest.approx(
        np.std(test_values.value()), rel=0.05
    ) == params[sigma3]["value"]
    assert pytest.approx(size, rel=0.005) == params[yield3]["value"]


def test_unbinned_simultaneous_nll():
    mu1, mu2, nll, sigma1, sigma2 = create_simultaneous_loss()
    minimizer = Minuit(tol=1e-5)
    status = minimizer.minimize(loss=nll, params=[mu1, sigma1, mu2, sigma2])
    params = status.params
    assert set(nll.get_params()) == {mu1, mu2, sigma1, sigma2}

    assert pytest.approx(np.mean(test_values_np), rel=0.007) == params[mu1]["value"]
    assert pytest.approx(np.mean(test_values_np2), rel=0.007) == params[mu2]["value"]
    assert pytest.approx(np.std(test_values_np), rel=0.007) == params[sigma1]["value"]
    assert pytest.approx(np.std(test_values_np2), rel=0.007) == params[sigma2]["value"]


@pytest.mark.flaky(3)
@pytest.mark.parametrize(
    "weights",
    (None, np.random.normal(loc=1.0, scale=0.2, size=test_values_np.shape[0])),
)
@pytest.mark.parametrize("sigma", (constr, constr_tf, covariance, covariance_tf))
@pytest.mark.parametrize("options", ({"subtr_const": False}, {"subtr_const": True}))
def test_unbinned_nll(weights, sigma, options):
    gaussian1, mu1, sigma1 = create_gauss1()
    gaussian2, mu2, sigma2 = create_gauss2()

    test_values = tf.constant(test_values_np)
    test_values = zfit.Data.from_tensor(obs=obs1, tensor=test_values, weights=weights)
    nll_object = zfit.loss.UnbinnedNLL(
        model=gaussian1, data=test_values, options=options
    )
    minimizer = Minuit(tol=1e-5)
    status = minimizer.minimize(loss=nll_object, params=[mu1, sigma1])
    params = status.params
    rel_error = 0.12 if weights is None else 0.1  # more fluctuating with weights

    assert pytest.approx(np.mean(test_values_np), rel=rel_error) == params[mu1]["value"]
    assert pytest.approx(
        np.std(test_values_np), rel=rel_error
    ) == params[sigma1]["value"]

    constraints = zfit.constraint.GaussianConstraint(
        params=[mu2, sigma2],
        observation=[mu_constr[0], sigma_constr[0]],
        cov=sigma(),
    )
    nll_object = UnbinnedNLL(
        model=gaussian2, data=test_values, constraints=constraints, options=options
    )

    minimizer = Minuit(tol=1e-4)
    status = minimizer.minimize(loss=nll_object, params=[mu2, sigma2])
    params = status.params
    if weights is None:
        assert params[mu2]["value"] > np.average(test_values_np, weights=weights)
        assert params[sigma2]["value"] < np.std(test_values_np)


def test_add():
    param1 = zfit.Parameter("param1", 1.0)
    param2 = zfit.Parameter("param2", 2.0)
    param3 = zfit.Parameter("param3", 2.0)

    pdfs = [0] * 4
    pdfs[0] = Gauss(param1, 4, obs=obs1)
    pdfs[1] = Gauss(param2, 5, obs=obs1)
    pdfs[2] = Gauss(3, 6, obs=obs1)
    pdfs[3] = Gauss(4, 7, obs=obs1)

    datas = [0] * 4
    datas[0] = zfit.Data.from_tensor(obs=obs1, tensor=z.constant(1.0))
    datas[1] = zfit.Data.from_tensor(obs=obs1, tensor=z.constant(2.0))
    datas[2] = zfit.Data.from_tensor(obs=obs1, tensor=z.constant(3.0))
    datas[3] = zfit.Data.from_tensor(obs=obs1, tensor=z.constant(4.0))

    ranges = [0] * 4
    ranges[0] = (1, 4)
    ranges[1] = Space(limits=(2, 5), obs=obs1)
    ranges[2] = Space(limits=(3, 6), obs=obs1)
    ranges[3] = Space(limits=(4, 7), obs=obs1)

    constraint1 = zfit.constraint.GaussianConstraint(
        params=param1, observation=1.0, sigma=0.5
    )
    constraint2 = zfit.constraint.GaussianConstraint(
        params=param3, observation=2.0, sigma=0.25
    )
    merged_contraints = [constraint1, constraint2]

    nll1 = UnbinnedNLL(
        model=pdfs[0], data=datas[0], fit_range=ranges[0], constraints=constraint1
    )
    nll2 = UnbinnedNLL(
        model=pdfs[1], data=datas[1], fit_range=ranges[1], constraints=constraint2
    )
    nll3 = UnbinnedNLL(
        model=[pdfs[2], pdfs[3]],
        data=[datas[2], datas[3]],
        fit_range=[ranges[2], ranges[3]],
    )

    simult_nll = nll1 + nll2 + nll3

    assert simult_nll.model == pdfs
    assert simult_nll.data == datas

    ranges[0] = Space(
        limits=ranges[0], obs="obs1", axes=(0,)
    )  # for comparison, Space can only compare with Space
    ranges[1].coords._axes = (0,)
    ranges[2].coords._axes = (0,)
    ranges[3].coords._axes = (0,)
    assert simult_nll.fit_range == ranges

    def eval_constraint(constraints):
        return z.reduce_sum([c.value() for c in constraints])

    assert eval_constraint(simult_nll.constraints) == eval_constraint(merged_contraints)
    assert set(simult_nll.get_params()) == {param1, param2, param3}


@pytest.mark.parametrize("numgrad", [True, False], ids=["numgrad", "autograd"])
@pytest.mark.parametrize("chunksize", [10000000, 1000])
def test_gradients(chunksize, numgrad):
    from numdifftools import Gradient

    zfit.run.chunking.active = True
    zfit.run.chunking.max_n_points = chunksize

    initial1 = 1.0
    initial2 = 2
    param1 = zfit.Parameter("param1", initial1)
    param2 = zfit.Parameter("param2", initial2)

    gauss1 = Gauss(param1, 4, obs=obs1)
    gauss2 = Gauss(param2, 5, obs=obs1)

    data1 = zfit.Data.from_tensor(obs=obs1, tensor=z.constant(1.0, shape=(100,)))
    data2 = zfit.Data.from_tensor(obs=obs1, tensor=z.constant(1.0, shape=(100,)))

    nll = UnbinnedNLL(model=[gauss1, gauss2], data=[data1, data2])

    def loss_funcparam1(values):
        with zfit.param.set_values(param2, values):
            return nll.value()

    def loss_funcparam1and2(values):
        with zfit.param.set_values([param2, param1], values):
            return nll.value()

    gradient_func_num = Gradient(loss_funcparam1)

    gradient1_num = gradient_func_num([param2])
    gradient1 = nll.gradient(params=param2, numgrad=numgrad)
    grad_num_truth = jacobi.jacobi(loss_funcparam1, [param2])[0]
    np.testing.assert_allclose(gradient1, grad_num_truth, rtol=1e-6)
    np.testing.assert_allclose(gradient1, gradient1_num, rtol=1e-6)
    param1.set_value(initial1)
    param2.set_value(initial2)
    params = [param2, param1]
    gradient2 = nll.gradient(params=params, numgrad=numgrad)
    gradient_func1and2 = Gradient(loss_funcparam1and2, order=2, base_step=0.1)
    gradient2_true_numdiff = gradient_func1and2([initial2, initial1])
    gradient2_true = jacobi.jacobi(loss_funcparam1and2, [initial2, initial1])[0]
    np.testing.assert_allclose(gradient2_true, gradient2_true_numdiff, rtol=1e-6)  # if this fails, numdiff/jacobi disagree
    np.testing.assert_allclose(gradient2, gradient2_true, rtol=1e-5)

    param1.set_value(initial1)
    param2.set_value(initial2)

    params3 = nll.get_params()
    gradient3 = nll.gradient(params3)
    gradients_true3 = []
    for param_o in params3:
        for param, grad in zip(params, gradient2):
            if param_o is param:
                gradients_true3.append(float(grad))
                break
    np.testing.assert_allclose(gradient3, gradients_true3, rtol=1e-5)


def test_simple_loss():
    true_a = 1.0
    true_b = 4.0
    true_c = -0.3
    truevals = true_a, true_b, true_c
    a_param = zfit.Parameter(
        "variable_a15151loss", 1.5, -1.0, 20.0, stepsize=z.constant(0.1)
    )
    b_param = zfit.Parameter("variable_b15151loss", 3.5)
    c_param = zfit.Parameter("variable_c15151loss", -0.23)
    param_list = [a_param, b_param, c_param]

    def loss_func(params):
        a_param, b_param, c_param = params
        probs = (
            z.convert_to_tensor(
                (a_param - true_a) ** 2
                + (b_param - true_b) ** 2
                + (c_param - true_c) ** 4
            )
            + 0.42
        )
        return tf.reduce_sum(input_tensor=tf.math.log(probs))

    with pytest.raises(ValueError):
        _ = zfit.loss.SimpleLoss(func=loss_func, params=param_list)

    loss_func.errordef = 1
    loss_deps = zfit.loss.SimpleLoss(func=loss_func, params=param_list)
    loss = zfit.loss.SimpleLoss(func=loss_func, params=param_list)
    loss2 = zfit.loss.SimpleLoss(func=loss_func, params=truevals)

    assert set(loss_deps.get_params()) == set(param_list)

    loss_tensor = loss_func(param_list)
    loss_value_np = loss_tensor

    assert pytest.approx(loss_value_np) == loss.value()
    assert pytest.approx(loss_value_np) == loss_deps.value()

    assert pytest.approx(
        loss_deps.value(full=True)
    ) == loss.value(full=True)

    minimizer = zfit.minimize.Minuit()
    result = minimizer.minimize(loss=loss)
    assert result.valid
    assert pytest.approx(result.params[a_param]["value"], rel=0.03) == true_a
    assert pytest.approx(result.params[b_param]["value"], rel=0.06) == true_b
    assert pytest.approx(result.params[c_param]["value"], rel=0.5) == true_c

    zfit.param.set_values(param_list, znp.asarray(param_list) + 0.6)
    result2 = minimizer.minimize(loss=loss2)
    assert result2.valid
    params = list(result2.params)
    assert pytest.approx(result2.params[params[0]]["value"], rel=0.03) == true_a
    assert pytest.approx(result2.params[params[1]]["value"], rel=0.06) == true_b
    assert pytest.approx(result2.params[params[2]]["value"], rel=0.5) == true_c


def test_simple_loss_addition():

    pa = zfit.Parameter("pa", 1.0)
    pb = zfit.Parameter("pb", 2.0)
    pc = zfit.Parameter("pc", 3.0)
    pd = zfit.Parameter("pd", 4.0)
    pe = zfit.Parameter("pe", 5.0)

    loss1 = zfit.loss.SimpleLoss(lambda x: x[0] ** 2, params=pa, errordef=0.5)
    loss2 = zfit.loss.SimpleLoss(lambda x: x[0] ** 2 + x[1] ** 2, params=[pb, pc], errordef=0.5)
    loss3 = zfit.loss.SimpleLoss(lambda x: x[0] ** 2 + x[1] ** 2 + x[2] ** 2, params=[pa, pb, pc], errordef=2)
    loss4 = zfit.loss.SimpleLoss(lambda x: x[0] ** 2 + x[1] ** 2, [pd, pe], errordef=0.5)

    loss12 = loss1 + loss2

    assert set(loss12.get_params()) == {pa, pb, pc}
    assert loss12.errordef == 0.5
    assert loss12.value() == loss1.value() + loss2.value()
    assert loss12.value(full=True) == loss1.value(full=True) + loss2.value(full=True)

    loss123 = loss1 + loss2 + loss3
    assert set(loss123.get_params()) == {pa, pb, pc}
    assert loss123.errordef == 0.5
    assert loss123.value() == loss1.value() + loss2.value() + 0.25 * loss3.value()
    assert loss123.value(full=True) == loss1.value(full=True) + loss2.value(full=True) + 0.25 * loss3.value(full=True)

    loss1234 = loss1 + loss2 + loss3 + loss4
    assert set(loss1234.get_params()) == {pa, pb, pc, pd, pe}
    assert loss1234.errordef == 0.5
    assert loss1234.value() == loss1.value() + loss2.value() + 0.25 * loss3.value() + loss4.value()
    assert loss1234.value(full=True) == loss1.value(full=True) + loss2.value(full=True) + 0.25 * loss3.value(full=True) + loss4.value(full=True)

    loss3124 = loss3 + loss1 + loss2 + loss4
    assert set(loss3124.get_params()) == {pa, pb, pc, pd, pe}
    assert loss3124.errordef == 2
    assert loss3124.value() ==  loss3.value() + 4 * loss1.value() + 4 * loss2.value() + 4 * loss4.value()
    assert loss3124.value(full=True) == loss3.value(full=True) + 4 * loss1.value(full=True) + 4 * loss2.value(full=True) + 4 * loss4.value(full=True)





def test_create_new_nll():
    gaussian1, mu1, sigma1 = create_gauss1()
    gaussian2, mu2, sigma2 = create_gauss2()

    test_values = tf.constant(test_values_np)
    test_values = zfit.Data.from_tensor(obs=obs1, tensor=test_values)
    nll = zfit.loss.UnbinnedNLL(model=gaussian1, data=test_values)

    nll2 = nll.create_new(model=gaussian2)
    assert nll2.data[0] is nll.data[0]
    assert nll2.constraints == nll.constraints
    assert nll2._options == nll._options

    nll3 = nll.create_new()
    assert nll3.data[0] is nll.data[0]
    assert nll3.constraints == nll.constraints
    assert nll3._options == nll._options

    nll4 = nll.create_new(options={})
    assert nll4.data[0] is nll.data[0]
    assert nll4.constraints == nll.constraints
    assert nll4._options != nll._constraints


def test_create_new_extnll():
    gaussian1, mu1, sigma1, yield1 = create_gauss3ext()

    test_values = tf.constant(test_values_np)
    test_values = zfit.Data.from_tensor(obs=obs1, tensor=test_values)
    nll = zfit.loss.ExtendedUnbinnedNLL(
        model=gaussian1,
        data=test_values,
        constraints=zfit.constraint.GaussianConstraint(mu1, 1.0, sigma=0.1),
    )

    nll2 = nll.create_new(model=gaussian1)
    assert nll2.data[0] is nll.data[0]
    assert nll2.constraints == nll.constraints
    assert nll2._options == nll._options

    nll3 = nll.create_new()
    assert nll3.data[0] is nll.data[0]
    assert nll3.constraints == nll.constraints
    assert nll3._options == nll._options

    nll4 = nll.create_new(options={})
    assert nll4.data[0] is nll.data[0]
    assert nll4.constraints == nll.constraints
    assert nll4._options != nll._constraints


def test_create_new_simple():
    _, mu1, sigma1 = create_gauss1()

    loss = zfit.loss.SimpleLoss(lambda x, y: x * y, params=[mu1, sigma1], errordef=0.5)

    loss1 = loss.create_new()
    assert loss1._simple_func is loss._simple_func
    assert loss1.errordef == loss.errordef


@pytest.mark.parametrize(
    "create_loss", [lambda: create_simultaneous_loss()[2], create_simpel_loss]
)
def test_callable_loss(create_loss):
    loss = create_loss()

    params = list(loss.get_params())
    x = np.array(znp.asarray(params)) + 0.1
    value_loss = loss(x)
    with zfit.param.set_values(params, x):
        true_val = zfit.run(loss.value(full=True))
        _ = zfit.run(loss.value(full=True))
        assert pytest.approx(value_loss) == true_val
        with pytest.raises(BehaviorUnderDiscussion):
            assert pytest.approx((loss())) == true_val

    with pytest.raises(ValueError):
        loss(x[:-1])
    with pytest.raises(ValueError):
        loss(list(x) + [1])


@pytest.mark.parametrize(
    "create_loss", [lambda: create_simultaneous_loss()[2], create_simpel_loss]
)
def test_iminuit_compatibility(create_loss):
    loss = create_loss()

    params = list(loss.get_params())
    x = znp.asarray(params) + 0.1
    zfit.param.set_values(params, x)

    with pytest.raises(ValueError):
        loss(x[:-1])
    with pytest.raises(ValueError):
        loss(list(x) + [1])

    import iminuit

    minimizer = iminuit.Minuit(loss, x)
    result = minimizer.migrad()
    assert result.valid
    minimizer.hesse()

    zfit.param.set_values(params, x)
    minimizer_zfit = zfit.minimize.Minuit()
    result_zfit = minimizer_zfit.minimize(loss)
    assert pytest.approx(result.fmin.fval, abs=0.03) == result_zfit.fmin


@pytest.mark.flaky(3)
@pytest.mark.parametrize(
    "weights", [None, np.random.normal(loc=1.0, scale=0.1, size=10000)]
)
# @pytest.mark.parametrize("weights", [None])
def test_binned_nll(weights):
    obs = zfit.Space("obs1", limits=(-15, 25))
    gaussian1, mu1, sigma1 = create_gauss1(obs=obs)
    gaussian2, mu2, sigma2 = create_gauss2(obs=obs)
    test_values_np = np.random.normal(loc=mu_true, scale=4, size=(10000, 1))

    test_values = tf.constant(test_values_np)
    test_values = zfit.Data.from_tensor(obs=obs, tensor=test_values, weights=weights)
    test_values_binned = test_values.to_binned(100)
    nll_object = zfit.loss.BinnedNLL(
        model=gaussian1.to_binned(test_values_binned.axes), data=test_values_binned
    )
    minimizer = Minuit()
    status = minimizer.minimize(loss=nll_object, params=[mu1, sigma1])
    params = status.params
    rel_error = 0.035 if weights is None else 0.15  # more fluctuating with weights

    assert pytest.approx(np.mean(test_values_np), rel=rel_error) == params[mu1]["value"]
    assert pytest.approx(
        np.std(test_values_np), rel=rel_error
    ) == params[sigma1]["value"]

    constraints = zfit.constraint.GaussianConstraint(
        params=[mu2, sigma2],
        observation=[mu_constr[0], sigma_constr[0]],
        uncertainty=[mu_constr[1], sigma_constr[1]],
    )
    nll_object = zfit.loss.BinnedNLL(
        model=gaussian2.to_binned(test_values_binned.axes),
        data=test_values_binned,
        constraints=constraints,
    )

    minimizer = Minuit()
    status = minimizer.minimize(loss=nll_object, params=[mu2, sigma2])
    params = status.params
    if weights is None:
        assert params[mu2]["value"] > np.mean(test_values_np)
        assert params[sigma2]["value"] < np.std(test_values_np)


def test_loss_from_data_outside_fails():

    obs = zfit.Space("obs1", limits=(-15, 25))
    pdf = zfit.pdf.Gauss(1.0, 1.0, obs=obs)
    data = np.linspace(-20, 20, 1000)

    with pytest.raises(IntentionAmbiguousError):
        zfit.loss.UnbinnedNLL(model=pdf, data=data)



    loss = zfit.loss.UnbinnedNLL(model=pdf, data=zfit.Data(data, obs=obs.obs))
    assert loss.value() ** 2 > -1  # just a sanity check
    loss2 = zfit.loss.UnbinnedNLL(model=pdf, data=zfit.Data(data, obs=obs))
    assert loss2.value() ** 2 > -1  # just a sanity check



#  Copyright (c) 2024 zfit
from __future__ import annotations

import tensorflow as tf

import zfit

# IMPORTANT! The communication of which axis corresponds to which data point happens here. So the user knows now that
# he should create this pdf with a space in the obs (x, y, z).


class CustomPDF(zfit.pdf.ZPDF):
    """3-dimensional PDF calculating doing something fancy, takes x, y, z as data."""

    _PARAMS = ("alpha", "beta")  # specify which parameters to take
    _N_OBS = 3

    @zfit.supports()
    def _unnormalized_pdf(self, x, params):  # implement function
        x0 = x[0]
        x1 = x[1]
        x2 = x[2]
        alpha = params["alpha"]
        beta = params["beta"]
        x0_new = tf.math.cos(alpha) * x0
        x1_new = tf.math.sinh(beta) * x1
        x2_new = x2 + 4.2
        return x0_new**2 + x1_new**2 + x2_new**2


xobs = zfit.Space("xobs", -4, 4)
yobs = zfit.Space("yobs", -3, 3)
zobs = zfit.Space("z", -2, 2)
obs = xobs * yobs * zobs

alpha = zfit.Parameter("alpha", 0.2)  # floating
beta = zfit.Parameter("beta", 0.4, floating=False)  # non-floating
custom_pdf = CustomPDF(obs=obs, alpha=alpha, beta=beta)

integral = custom_pdf.integrate(limits=obs)  # = 1 since normalized
sample = custom_pdf.sample(n=1000)  # DO NOT USE THIS FOR TOYS!
prob = custom_pdf.pdf(sample)  # DO NOT USE THIS FOR TOYS!



#  Copyright (c) 2024 zfit

from __future__ import annotations

from collections.abc import Mapping

import numpy as np

import zfit.z.numpy as znp

from ..core.interfaces import ZfitLoss
from ..core.parameter import Parameter, assign_values
from ..util.cache import GraphCachable
from ..util.exception import MaximumIterationReached
from .baseminimizer import BaseMinimizer, minimize_supports
from .fitresult import Approximations, FitResult
from .strategy import ZfitStrategy
from .termination import ConvergenceCriterion


class OptimizeStop(Exception):
    pass


class LevenbergMarquardt(BaseMinimizer, GraphCachable):
    _DEFAULT_name = "LM"

    def __init__(
        self,
        tol: float | None = None,
        mode: int | None = None,
        rho_min: float | None = None,
        rho_max: float | None = None,
        verbosity: int | None = None,
        options: Mapping[str, object] | None = None,
        maxiter: int | None = None,
        criterion: ConvergenceCriterion | None = None,
        strategy: ZfitStrategy | None = None,
        name: str | None = None,
    ):
        """Levenberg-Marquardt minimizer for general non-linear minimization by interpolating between Gauss-Newton and
        Gradient descent optimization.

        LM minimizes a function by iteratively solving a locally linearized
        version of the problem. Using the gradient (g) and the Hessian (H) of
        the loss function, the algorithm determines a step (h) that minimizes
        the loss function by solving :math:`Hh = g`. This works perfectly in one
        step for linear problems, however for non-linear problems it may be
        unstable far from the minimum. Thus a scalar damping parameter (L) is
        introduced and the Hessian is modified based on this damping. The form
        of the modification depends on the ``mode`` parameter, where the
        simplest example is :math:`H' = H + L \\mathbb{I}` (where
        :math:`\\mathbb{I}` is the identity). For a clarifying discussion of the
        LM algorithm see: Gavin, Henri 2016
        (https://msulaiman.org/onewebmedia/LM%20Method%20matlab%20codes%20and%20implementation.pdf)

        Args:
            tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
            mode: The mode of the LM algorithm. The default mode (0) is the
                simplest form of the LM algorithm. Other modes may be implemented
                in the future.
            rho_min: The minimum acceptable value of the ratio of the actual
                reduction in the loss function to the expected reduction. If the
                ratio is less than this value, the damping parameter is increased.
            rho_max: The maximum acceptable value of the ratio of the actual
                reduction in the loss function to the expected reduction. If the
                ratio is greater than this value, the damping parameter is
                decreased.
            verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizers offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
            options: |@doc:minimizer.options||@docend:minimizer.options|
            maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the ``value``, 'gradient`` or ``hessian``. |@docend:minimizer.maxiter|
            criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
            strategy: |@doc:minimizer.strategy| A class of type ``ZfitStrategy`` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
            name: |@doc:minimizer.name||@docend:minimizer.name

        """

        mode = 0 if mode is None else mode
        if mode != 0:
            msg = "Only mode 0 is currently implemented"
            raise NotImplementedError(msg)
        rho_min = 0.1 if rho_min is None else rho_min
        if rho_min <= 0:
            msg = "rho_min has to be > 0"
            raise ValueError(msg)
        rho_max = 2.0 if rho_max is None else rho_max
        if rho_max <= 1:
            msg = "rho_max has to be > 1"
            raise ValueError(msg)

        self.mode = mode
        self.rho_min = rho_min
        self.rho_max = rho_max
        self.Lup = 11
        self.Ldn = 9
        super().__init__(
            name=name,
            strategy=strategy,
            tol=tol,
            verbosity=verbosity,
            criterion=criterion,
            maxiter=maxiter,
            minimizer_options=options,
        )

    @staticmethod
    def _damped_hess0(hess, L):
        I = znp.eye(hess.shape[0])  # noqa: E741
        D = znp.ones_like(hess) - I
        return hess * (I + D / (1 + L)) + L * I * (1 + znp.diag(hess))

    def _mode0_step(self, loss, params, L):
        init_chi2, grad = loss.value_gradient(params)
        hess = loss.hessian(params)
        grad = grad[..., None]  # right shape for the solve
        best = (znp.zeros_like(params), init_chi2, L)
        scary_best = (None, init_chi2, L)
        direction = "none"
        nostep = True
        for _ in range(10):
            # Determine damped hessian
            damped_hess = self._damped_hess0(hess, L)
            # Solve for step with minimum chi^2
            h = znp.linalg.solve(damped_hess, -grad)

            # Calculate expected improvement in chi^2
            expected_improvement = -znp.transpose(h) @ damped_hess @ h - 2 * znp.transpose(h) @ grad

            # Calculate new chi^2
            params1 = params + h[:, 0]
            chi2 = loss.value(params1)
            # Skip if chi^2 is nan
            if not np.isfinite(chi2):
                L *= self.Lup
                if direction == "better":
                    break
                direction = "worse"
                continue

            # Keep track of any chi^2 improvement
            if chi2 <= scary_best[1]:
                scary_best = (h, chi2, L)

            # Check if chi^2 improved the expected amount
            rho = (init_chi2 - chi2) / expected_improvement
            if rho < self.rho_min or rho > self.rho_max:
                L *= self.Lup
                if L > 1e9 or direction == "better":
                    break
                direction = "worse"
                continue

            # Check for new best chi^2
            if chi2 < best[1]:
                best = (h, chi2, L)
                L /= self.Ldn
                nostep = False
                if L < 1e-9 or direction == "worse":
                    break
                direction = "better"
            elif chi2 >= best[1] and direction in ["worse", "none"] and L < 1e9:
                L *= self.Lup
                direction = "worse"
                continue
            else:  # chi2 >= best[1] and direction == "better"
                break

            # Break if step already very successful (> 10% improvement)
            if (init_chi2 - best[1]) / init_chi2 > 0.1:
                break

        if nostep:
            # use any improvement if a safe improvement could not be found
            if scary_best[0] is not None:
                return scary_best
            msg = "No step found"
            raise OptimizeStop(msg)
        return {"step": best, "hessian": damped_hess, "gradient": grad[:, 0]}

    @minimize_supports(init=True)
    def _minimize(self, loss: ZfitLoss, params: list[Parameter], init):
        if init:
            assign_values(params=params, values=init)

        evaluator = self.create_evaluator(loss=loss, params=params)
        criterion = self.create_criterion(loss=loss, params=params)
        loss_history = [evaluator.value(params)]
        L = 1.0
        success = False
        paramvals = znp.asarray(params)
        step = None
        approx = None
        iterator = range(self.get_maxiter(n=len(params)))
        for _niter in iterator:
            try:
                new_point = self._mode0_step(evaluator, paramvals, L)
                step = new_point["step"]
            except OptimizeStop:
                if self.verbosity >= 7:
                    print(f"OptimizeStop, Iteration {_niter}, loss: {loss_history[-1]}")
                break
            L = step[2]
            loss_history.append(step[1])

            paramvals += step[0][:, 0]
            approx = Approximations(params=params, gradient=new_point.get("gradient"), hessian=new_point.get("hessian"))
            tempres = FitResult(
                loss=loss,
                params=dict(zip(params, paramvals)),
                minimizer=self,
                valid=False,
                criterion=criterion,
                edm=-999,
                fminopt=loss_history[-1],
                approx=approx,
            )
            success = criterion.converged(tempres)
            if self.verbosity >= 7:
                print(
                    f"Iteration {_niter}, loss: {loss_history[-1]}, success: {success}, criterion: {criterion.last_value}"
                )
            if success:
                if self.verbosity >= 6:
                    print(f"Converged with criterion: {criterion.last_value} < {self.tol}")
                break
        else:
            msg = f"Maximum number of iterations ({self.maxiter}) reached."
            raise MaximumIterationReached(msg)

        msg = "No step taken. This should not happen?"
        assert step is not None, msg
        assign_values(params, paramvals)

        return FitResult(
            loss=loss,
            params={p: float(p.value()) for p in params},
            edm=criterion.last_value,
            fminopt=loss_history[-1],
            minimizer=self,
            valid=success,
            criterion=criterion,
            approx=approx,
            evaluator=evaluator,
        )



#  Copyright (c) 2024 zfit
from __future__ import annotations

import numpy as np
import tensorflow as tf
from dotmap import DotMap

from .util.execution import RunManager

run = RunManager()


def set_seed(seed=None, numpy=None, backend=None, zfit=None):
    """Set random seed for zfit, numpy and the backend. If seed is given, this is used to generate deterministic seeds
    for numpy, zfit and the backend.

    Notably, no python random seed is set.

    .. warning::

        To use this function to guarantee randomness, do *not* specify a seed.
        Use a seed _only_ for reproducibility: there have been unique cases of correlation between the seed and the result.
        As the seeds are returned, you can save them and rerun *if needed*.

    Uses `os.urandom` to generate a seed if `None` is given, which guarantees cryptographically secure randomness.

    Args:
        seed (int, optional): Seed to set for the random number generator of the seeds for within zfit. If `None` (default), the seed is set to a random value.
        numpy (int, bool, optional): Seed to set for numpy. If `True` (default), a random seed depending on the seed as used for zfit is used. If `False`, the seed is not set.
        backend (int, bool, optional): Seed to set for the backend. If `True` (default), a random seed depending on the seed as used for zfit is used. If `False`, the seed is not set.
        zfit (int, bool, optional): Seed to set for zfit. If `True` (default), a random seed is used. If `False`, the seed is not set.

    Returns:
        dict: Seeds that were set, with the keys `zfit`, `numpy`, `backend`.
    """
    if seed is None:
        seed = generate_urandom_seed()

    if numpy is None:
        numpy = True
    if backend is None:
        backend = True
    if zfit is None:
        zfit = True
    initial_seed = seed

    if numpy is True:
        if seed is True:
            numpy = generate_urandom_seed()
        else:
            rng = np.random.default_rng(seed)
            seed = rng.integers(0, 2**31 - 1)
            numpy = seed
    if backend is True:
        if seed is True:
            backend = generate_urandom_seed()
        else:
            rng = np.random.default_rng(seed)
            seed = rng.integers(0, 2**31 - 1)
            backend = seed

    if numpy is not None and numpy is not False:
        np.random.seed(numpy)
    if backend is not None and backend is not False:
        tf.random.set_seed(backend)

    from .z.random import get_prng

    if zfit is True:
        if seed is True:
            zfit = generate_urandom_seed()
        else:
            rng = np.random.default_rng(seed)
            seed = rng.integers(0, 2**31 - 1)
            zfit = seed
    if zfit is not None and zfit is not False:
        get_prng().reset_from_seed(zfit)
    seeds = {"zfit": zfit, "numpy": numpy, "backend": backend}
    if initial_seed is not True:
        seeds["seed"] = initial_seed
    return seeds


def generate_urandom_seed():
    import os

    random_data = os.urandom(4)
    backend_seed = int.from_bytes(random_data, byteorder="big")
    return int(abs(backend_seed) % 2**31)  # make sure it's positive and not too large


_verbosity = 0


def set_verbosity(verbosity):
    global _verbosity
    _verbosity = verbosity


def get_verbosity():
    return _verbosity


ztypes = DotMap(
    {
        "float": tf.float64,
        "complex": tf.complex128,
        "int": tf.int64,
        "auto_upcast": True,
    }
)
upcast_ztypes = {
    tf.float16: tf.float64,
    tf.float32: tf.float64,
    tf.float64: tf.float64,
    tf.complex64: tf.complex128,
    tf.complex128: tf.complex128,
    tf.int8: tf.int64,
    tf.int16: tf.int64,
    tf.int32: tf.int64,
    tf.int64: tf.int64,
}

options = DotMap(
    {
        "epsilon": 1e-8,
        "numerical_grad": None,
        "advanced_warning": True,
        "changed_warning": True,
        "auto_update_params": True,
    }
)

advanced_warnings = DotMap(
    {
        "sum_extended_frac": True,
        "exp_shift": True,
        "py_func_autograd": True,
        "inconsistent_fitrange": True,
        "extended_in_UnbinnedNLL": True,
        "all": True,
    }
)

changed_warnings = DotMap(
    {
        "new_sum": True,
        "all": True,
    }
)



#  Copyright (c) 2024 zfit
import numpy as np
import pytest
import zfit
from zfit import z
import zfit.z.numpy as znp
import tensorflow as tf
from scipy import integrate

mu_true = 0.5
sigmal_true = 1.1
sigma_true = 1.5


def _numpy_bifurgauss_pdf(x, mu, sigmal, sigmar):
    norm = np.sqrt(2 / np.pi) / (sigmal + sigmar)
    if x < mu:
        return norm * np.exp(-0.5 * ((x - mu) / sigmal) ** 2)
    else:
        return norm * np.exp(-0.5 * ((x - mu) / sigmar) ** 2)

numpy_bifurgauss_pdf = np.vectorize(_numpy_bifurgauss_pdf, excluded=("mu", "sigmal", "sigmar"))


def create_bifurgauss(mu, sigmal, sigmar, limits):
    obs = zfit.Space("obs1", limits)
    bifurgauss = zfit.pdf.BifurGauss(mu=mu, sigmal=sigmal, sigmar=sigmar, obs=obs)
    return bifurgauss, obs



def test_bifurgauss_pdf():
    bifurgauss, obs = create_bifurgauss(mu=mu_true, sigmal=sigmal_true, sigmar=sigma_true, limits=(-5, 5))
    assert pytest.approx(
        np.atleast_1d(numpy_bifurgauss_pdf(0.5, mu=mu_true, sigmal=sigmal_true, sigmar=sigma_true)), rel=1e-5
    ) == bifurgauss.pdf(0.5, norm=False)
    test_values = znp.linspace(-5, 5, 10_000)
    np.testing.assert_allclose(
        bifurgauss.pdf(test_values, norm=False),
        numpy_bifurgauss_pdf(test_values, mu=mu_true, sigmal=sigmal_true, sigmar=sigma_true),
        rtol=1e-5,
    )
    assert np.all(bifurgauss.pdf(test_values, norm=False) <= bifurgauss.pdf(0.5, norm=False))

    sample = bifurgauss.sample(1000)
    assert np.all(np.isfinite(sample.value())), "Some samples from the bifurgauss PDF are NaN or infinite"
    assert sample.n_events == 1000
    assert np.all(tf.logical_and(-5 <= sample.value(), sample.value() <= 5))



def test_bifurgauss_integral():
    bifurgauss, obs = create_bifurgauss(mu=mu_true, sigmal=sigmal_true, sigmar=sigma_true, limits=(-5, 5))
    full_interval_analytic = bifurgauss.analytic_integrate(obs, norm=False)
    full_interval_numeric = bifurgauss.numeric_integrate(obs, norm=False)
    true_integral = 0.998442
    numpy_full_integral = integrate.quad(
            numpy_bifurgauss_pdf, -5, 5, args=(mu_true, sigmal_true, sigma_true)
        )[0]
    assert pytest.approx(true_integral, 1e-6) == full_interval_analytic
    assert pytest.approx(true_integral, 1e-6) == full_interval_numeric
    assert pytest.approx(numpy_full_integral, 1e-6) == full_interval_analytic
    assert pytest.approx(numpy_full_integral, 1e-6) == full_interval_numeric

    analytic_integral = bifurgauss.analytic_integrate(limits=(-1, 1), norm=False)
    numeric_integral = bifurgauss.numeric_integrate(limits=(-1, 1), norm=False)
    numpy_integral = integrate.quad(
        numpy_bifurgauss_pdf, -1, 1, args=(mu_true, sigmal_true, sigma_true)
    )[0]
    assert pytest.approx(numeric_integral, 1e-6) == analytic_integral
    assert pytest.approx(numpy_integral, 1e-6) == analytic_integral


def test_equivalency_with_generalizedcb():
    bifurgauss, obs = create_bifurgauss(mu=mu_true, sigmal=sigmal_true, sigmar=sigma_true, limits=(-5, 5))
    generalized_cb = zfit.pdf.GeneralizedCB(mu=mu_true, sigmal=sigmal_true, alphal=100, nl=1, sigmar=sigma_true, alphar=100, nr=1, obs=obs)

    assert pytest.approx(generalized_cb.pdf(0.5), rel=1e-5) == bifurgauss.pdf(0.5)
    test_values = znp.linspace(-5, 5, 10_000)
    np.testing.assert_allclose(
        bifurgauss.pdf(test_values),
        generalized_cb.pdf(test_values),
        rtol=1e-5,
    )

    assert pytest.approx(generalized_cb.analytic_integrate(obs), rel=1e-5) == bifurgauss.analytic_integrate(obs)
    assert pytest.approx(generalized_cb.numeric_integrate(obs), rel=1e-5) == bifurgauss.numeric_integrate(obs)
    assert pytest.approx(generalized_cb.analytic_integrate(limits=(-1, 1)), rel=1e-5) == bifurgauss.analytic_integrate(limits=(-1, 1))
    assert pytest.approx(generalized_cb.numeric_integrate(limits=(-1, 1)), rel=1e-5) == bifurgauss.numeric_integrate(limits=(-1, 1))



#  Copyright (c) 2024 zfit
import numpy as np
import pytest


@pytest.mark.parametrize("weights", [None, np.random.uniform(size=1000)], ids=['no_weights', 'weights'])
def test_update_data_sampler(weights):
    import zfit
    from zfit.data import SamplerData
    sample = np.random.normal(0, 1, size=(1000, 3))
    sample2 = np.random.normal(0, 1, size=(1000, 3))
    sample3 = np.random.normal(0, 1, size=(1000, 3))

    obs = zfit.Space('obs1', limits=(-100, 100)) * zfit.Space('obs2', limits=(-200, 200)) * zfit.Space('obs3', limits=(-150, 150))
    data = SamplerData.from_sampler(obs=obs, sample_and_weights_func=lambda n, params: (sample, weights), n=1000)
    if weights is not None:
        assert np.allclose(data.weights, weights)
    assert np.allclose(data.value(), sample)
    if weights is not None:
        with pytest.raises(ValueError):
            data.update_data(sample2, weights=None)
    data.update_data(sample2, weights=weights)
    assert np.allclose(data.value(), sample2)
    if weights is not None:
        with pytest.raises(ValueError):
            data.update_data(sample3, weights=None)
    else:
        data.update_data(sample3, weights=weights)
        assert np.allclose(data.value(), sample3)



#  Copyright (c) 2024 zfit

from __future__ import annotations

from typing import Callable, Literal

import pydantic.v1 as pydantic
import tensorflow as tf

import zfit.z.numpy as znp
from zfit.util.exception import AnalyticGradientNotAvailable

from ..core.serialmixin import SerializableMixin
from ..core.space import supports
from ..serialization import Serializer  # noqa: F401
from ..settings import ztypes
from ..util import ztyping
from .basefunctor import FunctorPDFRepr
from .functor import BaseFunctor


def get_value(cache: tf.Variable, flag: tf.Variable, func: Callable):
    @tf.custom_gradient
    def actual_func():
        def autoset_func():
            val = func()
            return cache.assign(val, read_value=True)

        def use_cache():
            return cache

        val = tf.cond(flag, use_cache, autoset_func)

        def grad_fn(dval, variables):  # noqa: ARG001
            msg = (
                "The analytic gradient is not implemented for caching PDF. Use the numerical gradient instead."
                "(either using zfit.run.set_autograd_mode(False) and/or by using the minimizer internal numerical gradient)"
            )
            raise AnalyticGradientNotAvailable(msg)

        return val, grad_fn

    return actual_func()


class CachedPDF(BaseFunctor, SerializableMixin):
    def __init__(
        self,
        pdf: ztyping.PDFInputType,
        *,
        extended: ztyping.ExtendedInputType = None,
        norm: ztyping.NormInputType = None,
        cache_tol=None,
        name: str | None = None,
        label: str | None = None,
    ):
        """Creates a PDF where ``pdf`` and ``integrate`` methods are cacheable.

        .. note::

           Analytic gradients are not available for the cached PDF. Use the numerical gradient instead,
           either by using a minimizers internal calculator (e.g. :py:class:`~zfit.minimize.Minuit(..., gradient=True)`) or by
           setting the autograd mode to False (e.g. :py:func:`~zfit.run.set_autograd_mode(False)`).
           An error will be raised if the analytic gradient is requested.

        The method stores the last calculated value of a function for a specific dataset and
        returns it when the input data and the parameters are the same. This can be useful when
        the pdf is called multiple times with the same data and parameters, for example in the
        minimization process when a numerical gradient is used.

        Args:
            pdf: pdf which methods to be cached
            cache_tol: accuracy of absolute tolerance comparing arguments (parameters, data) with cached values
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        obs = pdf.space
        hs3_init = {
            "obs": obs,
            "extended": extended,
            "norm": norm,
            "cache_tol": cache_tol,
            "name": name,
        }
        name = name or pdf.name
        super().__init__(pdfs=pdf, obs=obs, name=name, extended=extended, norm=norm, label=label, autograd_params=[])
        params = list(pdf.get_params(floating=None, is_yield=None, extract_independent=True))

        param_cache = tf.Variable(
            znp.zeros(shape=tf.shape(tf.stack(params)), dtype=ztypes.float),
            trainable=False,
            validate_shape=False,
            dtype=tf.float64,
        )
        param_cache_int = tf.Variable(
            znp.zeros(shape=tf.shape(tf.stack(params)), dtype=ztypes.float),
            trainable=False,
            validate_shape=False,
            dtype=tf.float64,
        )

        self._cached_pdf_params = param_cache
        self._cached_pdf_params_for_integration = param_cache_int
        self._pdf_cache = None
        self._cached_x = None
        self._pdf_cache_valid = tf.Variable(initial_value=False, trainable=False)

        self._cached_integral_limits = None
        self._integral_cache = None
        self._integral_cache_valid = tf.Variable(initial_value=False, trainable=False)

        self._cache_tolerance = 1e-8 if cache_tol is None else cache_tol
        self.hs3.original_init.update(hs3_init)

    @supports(norm="space")
    def _pdf(self, x, norm):
        x = x.value()
        if self._pdf_cache is None:
            self._pdf_cache = tf.Variable(
                -999.0
                * znp.ones(
                    shape=tf.shape(x)[0],
                ),  # negative ones, to make sure these are unrealistic values
                trainable=False,
                validate_shape=False,
                dtype=ztypes.float,
            )

        if self._cached_x is None:
            self._cached_x = tf.Variable(
                x + 19.0,  # to make sure it's not the same
                trainable=False,
                validate_shape=False,
                dtype=ztypes.float,
            )
        x_same = tf.math.reduce_all(znp.abs(x - self._cached_x) < self._cache_tolerance)
        pdf_params = list(self.pdfs[0].get_params())
        if hasparams := len(pdf_params) > 0:
            stacked_pdf_params = tf.stack(pdf_params)
            params_same = tf.math.reduce_all(
                znp.abs(stacked_pdf_params - self._cached_pdf_params) < self._cache_tolerance
            )

            same_args = tf.math.logical_and(params_same, x_same)
        else:
            same_args = x_same
        assign1 = self._pdf_cache_valid.assign(same_args, read_value=False)

        def value_update_func():
            if hasparams:
                self._cached_pdf_params.assign(stacked_pdf_params, read_value=False)
            self._cached_x.assign(x, read_value=False)
            return self.pdfs[0].pdf(x, norm)

        with tf.control_dependencies([assign1]):
            return get_value(self._pdf_cache, self._pdf_cache_valid, value_update_func)

    @supports(norm="space")
    def _integrate(self, limits, norm, options=None):
        if self._cached_integral_limits is None:
            self._cached_integral_limits = tf.Variable(
                tf.stack(limits.v1.limits) + 19.0,  # to make sure it's not the same
                trainable=False,
                validate_shape=False,
                dtype=ztypes.float,
            )

        if self._integral_cache is None:
            self._integral_cache = tf.Variable(
                znp.zeros(shape=tf.shape([1])),
                trainable=False,
                validate_shape=False,
                dtype=ztypes.float,
            )

        stacked_integral_limits = tf.stack(limits.v1.limits)
        limits_same = tf.math.reduce_all(
            znp.abs(stacked_integral_limits - self._cached_integral_limits) < self._cache_tolerance
        )

        params = list(self.pdfs[0].get_params(floating=None))
        if hasparams := len(params) > 0:
            stacked_pdf_params = tf.stack(params)
            params_same = tf.math.reduce_all(
                znp.abs(stacked_pdf_params - self._cached_pdf_params_for_integration) < self._cache_tolerance
            )

            same_args = tf.math.logical_and(params_same, limits_same)
        else:
            same_args = limits_same
        assign1 = self._integral_cache_valid.assign(same_args, read_value=False)

        def value_update_func():
            if hasparams:
                self._cached_pdf_params_for_integration.assign(stacked_pdf_params, read_value=False)
            self._cached_integral_limits.assign(stacked_integral_limits, read_value=False)
            return self.pdfs[0].integrate(limits, norm, options=options)

        with tf.control_dependencies([assign1]):
            return get_value(self._integral_cache, self._integral_cache_valid, value_update_func)


class CachedPDFRepr(FunctorPDFRepr):
    _implementation = CachedPDF
    hs3_type: Literal["CachedPDF"] = pydantic.Field("CachedPDF", alias="type")

    def _to_orm(self, init) -> CachedPDF:
        init.pop("obs")
        init["pdf"] = init.pop("pdfs")[0]
        return super()._to_orm(init)



#  Copyright (c) 2024 zfit
from __future__ import annotations

import numpy as np

import zfit

# create a multidimensional space
xobs = zfit.Space("xobs", -4, 4)
yobs = zfit.Space("yobs", -3, 5)
zobs = zfit.Space("z", -2, 4)
obs = xobs * yobs * zobs
assert obs.n_obs == 3

# retrieve a subspace
xobs = obs.with_obs("xobs")
yxobs = obs.with_obs(["yobs", "xobs"])  # note the change in order

# retrieve the limits
# they have been overcomplicated in the past, now they are simple. Just access the limits via space.v1
limits = obs.v1.limits  # (lower, upper)
lower = obs.v1.lower
upper = obs.v1.upper

assert np.all(lower == [-4, -3, -2])
assert np.all(upper == [4, 5, 4])

# the volume is the product of the volumes of the individual spaces
volume = obs.volume
assert volume == xobs.volume * yobs.volume * zobs.volume
assert xobs.volume == xobs.v1.upper - xobs.v1.lower

# for example, creating a linspace object is simple
x = np.linspace(xobs.v1.lower, xobs.v1.upper, 1000)
# or
x = np.linspace(*xobs.v1.limits, 1000)

# or even in 3D
x = np.linspace(*obs.v1.limits, 1000)
assert x.shape == (1000, 3)



#  Copyright (c) 2024 zfit
"""Exceptions that are raised by zfit.

Some are to provide a more specific error message, others are to steer the execution by raising an error that will be
caught in the right place.
"""

from __future__ import annotations

from .core.serialmixin import NumpyArrayNotSerializableError
from .util.checks import NONE
from .util.deprecation import deprecated
from .util.exception import (
    AnalyticGradientNotAvailable,
    AnalyticIntegralNotImplemented,
    AnalyticNotImplemented,
    AnalyticSamplingNotImplemented,
    BreakingAPIChangeError,
    FunctionNotImplemented,
    IllegalInGraphModeError,
    InitNotImplemented,
    LogicalUndefinedOperationError,
    MaximumIterationReached,
    MinimizerSubclassingError,
    MultipleLimitsNotImplemented,
    NormRangeNotImplemented,
    ParamNameNotUniqueError,
    ShapeIncompatibleError,
    SpecificFunctionNotImplemented,
    VectorizedLimitsNotImplemented,
)

__all__ = [
    "NameAlreadyTakenError",
    "ParamNameNotUniqueError",
    "IllegalInGraphModeError",
    "NormRangeNotImplemented",
    "MultipleLimitsNotImplemented",
    "VectorizedLimitsNotImplemented",
    "AnalyticNotImplemented",
    "FunctionNotImplemented",
    "AnalyticSamplingNotImplemented",
    "AnalyticIntegralNotImplemented",
    "SpecificFunctionNotImplemented",
    "MinimizerSubclassingError",
    "MaximumIterationReached",
    "ShapeIncompatibleError",
    "InitNotImplemented",
    "LogicalUndefinedOperationError",
    "NumpyArrayNotSerializableError",
    "AnalyticGradientNotAvailable",
    "IpyoptPicklingError",
    "NONE",
]


class VectorizedLimitsNotImplementedError(VectorizedLimitsNotImplemented):
    @deprecated(None, "Use VectorizedLimitsNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class SpecificFunctionNotImplementedError(SpecificFunctionNotImplemented):
    @deprecated(None, "Use SpecificFunctionNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class NormRangeNotImplementedError(NormRangeNotImplemented):
    @deprecated(None, "Use NormRangeNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class MultipleLimitsNotImplementedError(MultipleLimitsNotImplemented):
    @deprecated(None, "Use MultipleLimitsNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class FunctionNotImplementedError(FunctionNotImplemented):
    @deprecated(None, "Use FunctionNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class AnalyticSamplingNotImplementedError(AnalyticSamplingNotImplemented):
    @deprecated(None, "Use AnalyticSamplingNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class AnalyticIntegralNotImplementedError(AnalyticIntegralNotImplemented):
    @deprecated(None, "Use AnalyticIntegralNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class AnalyticNotImplementedError(AnalyticNotImplemented):
    @deprecated(None, "Use AnalyticNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class NameAlreadyTakenError(Exception):
    def __init__(self, *_: object) -> None:
        raise BreakingAPIChangeError(
            msg="NameAlreadyTakenError has been removed and the behavior has substantially changed:"
            "parameters are now allowed to exist with the same as long as they are not within the same PDF/loss/func."
        )


class IpyoptPicklingError(TypeError):
    pass


class OutsideLimitsError(Exception):
    pass


class AutogradNotSupported(Exception):
    pass



#  Copyright (c) 2024 zfit

from __future__ import annotations

from collections.abc import Callable, Iterable

import tensorflow as tf

import zfit.z.numpy as znp

from ..core.basefunc import BaseFuncV1
from ..core.basemodel import SimpleModelSubclassMixin
from ..core.interfaces import ZfitFunc
from ..core.space import supports
from ..models.basefunctor import FunctorMixin
from ..util import ztyping
from ..util.container import convert_to_container


class SimpleFuncV1(BaseFuncV1):
    def __init__(
        self,
        obs: ztyping.ObsTypeInput,
        func: Callable,
        name: str = "Function",
        **params,
    ):
        """Create a simple function out of of `func` with the observables `obs` depending on `parameters`.

        Args:
            func:
            obs:
            name:
            **params: The parameters as keyword arguments. E.g. `mu=Parameter(...)`
        """
        super().__init__(name=name, obs=obs, params=params)
        self._value_func = self._check_input_x_function(func)

    def _func(self, x):
        try:
            return self._value_func(x)
        except TypeError:  # self requested, TODO maybe check signature?
            return self._value_func(self, x)


class BaseFunctorFuncV1(FunctorMixin, BaseFuncV1):
    def __init__(self, funcs, name="BaseFunctorFunc", params=None, **kwargs):
        funcs = convert_to_container(funcs)
        if params is None:
            params = {}
        # for func in funcs:
        #     params.update(func.params)

        self.funcs = funcs
        super().__init__(name=name, models=self.funcs, params=params, **kwargs)
        self._models = self.funcs


class SumFunc(BaseFunctorFuncV1):
    def __init__(
        self,
        funcs: Iterable[ZfitFunc],
        obs: ztyping.ObsTypeInput = None,
        name: str = "SumFunc",
        **kwargs,
    ):
        super().__init__(funcs=funcs, obs=obs, name=name, **kwargs)

    def _func(self, x):
        # sum_funcs = tf.add_n([func.value(x) for func in self.funcs])
        funcs = [func.func(x) for func in self.funcs]
        return tf.math.accumulate_n(funcs)

    @supports()
    def _analytic_integrate(self, limits, norm):
        # below may raises AnalyticIntegralNotImplementedError, that's fine. We don't wanna catch that.
        integrals = [func.analytic_integrate(limits=limits, norm=norm) for func in self.funcs]
        return tf.math.accumulate_n(integrals)


class ProdFunc(BaseFunctorFuncV1):
    def __init__(
        self,
        funcs: Iterable[ZfitFunc],
        obs: ztyping.ObsTypeInput = None,
        name: str = "SumFunc",
        **kwargs,
    ):
        super().__init__(funcs=funcs, obs=obs, name=name, **kwargs)

    def _func(self, x):
        funcs = [func.func(x) for func in self.funcs]
        return znp.prod(funcs, axis=0)


class ZFuncV1(SimpleModelSubclassMixin, BaseFuncV1):
    def __init__(self, obs: ztyping.ObsTypeInput, name: str = "ZFunc", **params):
        super().__init__(obs=obs, name=name, **params)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._check_simple_model_subclass()



#!/usr/bin/env python
#  Copyright (c) 2024 zfit
from __future__ import annotations

import argparse
import os
import re
from pathlib import Path

import yaml

here = Path(os.path.realpath(__file__)).parent

parser = argparse.ArgumentParser(
    description="Replace arguments with central stored ones",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)

parser.add_argument("files", nargs="*", help="Files to be processed.")

parser.add_argument("--dry", action="store_true", help="Dry run WITHOUT replacing.")

cfg = parser.parse_args()

with Path(here / "argdocs.yaml").open() as replfile:
    replacements = yaml.load(replfile, Loader=yaml.Loader)

# Replace the target string
# auto_end_old = r'|@docend|'
for filepath in cfg.files:
    if not filepath.endswith(".py"):
        continue
    with Path(filepath).open() as file:
        filedata = file.read()

    infile = False
    needs_replacement = False
    for param, replacement in replacements.items():
        replacement = replacement.rstrip("\n")
        while replacement[:1] == " ":  # we want to remove the whitespace
            replacement = replacement[1:]
        auto_start = rf"|@doc:{param}|"
        auto_end = rf"|@docend:{param}|"
        matches = re.findall(
            auto_start.replace("|", r"\|") + r".*?" + auto_end.replace("|", r"\|"),
            filedata,
            re.DOTALL,
        )

        if not matches:
            continue
        infile = True

        replacement_mod = f"{auto_start} {replacement} {auto_end}"

        for match in matches:
            if auto_start in match[len(auto_start) :]:  # sanity check
                msg = f"Docstring formatting error," f" has more than one start until an end command: {match}"
                raise ValueError(msg)
            if match != replacement_mod:
                needs_replacement = True
                filedata = filedata.replace(match, replacement_mod)

    # Write the file out again
    replace_msg = "replaced docs" if needs_replacement else "docs already there"
    filename = filepath.split("/")[-1]
    if infile:
        if cfg.dry:
            pass
        elif needs_replacement:
            with Path(filepath).open("w") as file:
                file.write(filedata)



#  Copyright (c) 2024 zfit
from __future__ import annotations

from .minimizers.baseminimizer import (
    BaseMinimizer,
    DefaultStrategy,
    minimize_supports,
)
from .minimizers.ipopt import Ipyopt
from .minimizers.minimizer_lm import LevenbergMarquardt
from .minimizers.minimizer_minuit import Minuit
from .minimizers.minimizer_nlopt import (
    NLoptBaseMinimizer,
    NLoptBOBYQA,
    NLoptCCSAQ,
    NLoptCOBYLA,
    NLoptESCH,
    NLoptISRES,
    NLoptLBFGS,
    NLoptMLSL,
    NLoptMMA,
    NLoptShiftVar,
    NLoptSLSQP,
    NLoptStoGO,
    NLoptSubplex,
    NLoptTruncNewton,
)
from .minimizers.minimizers_scipy import (
    ScipyBaseMinimizer,
    ScipyBFGS,
    ScipyCOBYLA,
    ScipyDogleg,
    ScipyLBFGSB,
    ScipyNelderMead,
    ScipyNewtonCG,
    ScipyPowell,
    ScipySLSQP,
    ScipyTruncNC,
    ScipyTrustConstr,
    ScipyTrustKrylov,
    ScipyTrustNCG,
)
from .minimizers.optimizers_tf import Adam, WrapOptimizer
from .minimizers.strategy import DefaultToyStrategy, PushbackStrategy
from .minimizers.termination import EDM

ScipyTrustConstrV1 = ScipyTrustConstr
ScipyTrustNCGV1 = ScipyTrustNCG
ScipyTrustKrylovV1 = ScipyTrustKrylov
ScipyDoglegV1 = ScipyDogleg
ScipyCOBYLAV1 = ScipyCOBYLA
ScipyLBFGSBV1 = ScipyLBFGSB
ScipyPowellV1 = ScipyPowell
ScipySLSQPV1 = ScipySLSQP
ScipyNewtonCGV1 = ScipyNewtonCG
ScipyTruncNCV1 = ScipyTruncNC
ScipyNelderMeadV1 = ScipyNelderMead
NLoptLBFGSV1 = NLoptLBFGS
NLoptTruncNewtonV1 = NLoptTruncNewton
NLoptSLSQPV1 = NLoptSLSQP
NLoptMMAV1 = NLoptMMA
NLoptCCSAQV1 = NLoptCCSAQ
NLoptShiftVarV1 = NLoptShiftVar
NLoptMLSLV1 = NLoptMLSL
NLoptStoGOV1 = NLoptStoGO
NLoptESCHV1 = NLoptESCH
NLoptISRESV1 = NLoptISRES
NLoptSubplexV1 = NLoptSubplex
NLoptBOBYQAV1 = NLoptBOBYQA
NLoptCOBYLAV1 = NLoptCOBYLA
IpyoptV1 = Ipyopt
ScipyBaseMinimizerV1 = ScipyBaseMinimizer
NLoptBaseMinimizerV1 = NLoptBaseMinimizer

BaseMinimizerV1 = BaseMinimizer


__all__ = [
    # temp added
    "ScipyDogleg",
    "ScipyCOBYLA",
    "ScipyTrustNCG",
    "ScipyTrustKrylov",
    # temp added end
    "WrapOptimizer",
    "Adam",
    "Minuit",
    "LevenbergMarquardt",
    "ScipyBaseMinimizerV1",
    "ScipyLBFGSBV1",
    "ScipyTrustConstrV1",
    "ScipyPowellV1",
    "ScipySLSQPV1",
    "ScipyNewtonCGV1",
    "ScipyTruncNCV1",
    "ScipyNelderMeadV1",
    "ScipyBaseMinimizer",
    "ScipyLBFGSB",
    "ScipyTrustConstr",
    "ScipyPowell",
    "ScipySLSQP",
    "ScipyNewtonCG",
    "ScipyTruncNC",
    "ScipyNelderMead",
    "ScipyBFGS",
    "NLoptBaseMinimizerV1",
    "NLoptLBFGSV1",
    "NLoptTruncNewtonV1",
    "NLoptSLSQPV1",
    "NLoptMMAV1",
    "NLoptCCSAQV1",
    "NLoptShiftVarV1",
    "NLoptMLSLV1",
    "NLoptStoGOV1",
    "NLoptESCHV1",
    "NLoptISRESV1",
    "NLoptSubplexV1",
    "NLoptBOBYQAV1",
    "NLoptCOBYLAV1",
    "NLoptBaseMinimizer",
    "NLoptLBFGS",
    "NLoptTruncNewton",
    "NLoptSLSQP",
    "NLoptMMA",
    "NLoptCCSAQ",
    "NLoptShiftVar",
    "NLoptMLSL",
    "NLoptStoGO",
    "NLoptESCH",
    "NLoptISRES",
    "NLoptSubplex",
    "NLoptBOBYQA",
    "NLoptCOBYLA",
    "Ipyopt",
    "LevenbergMarquardt",
    "BaseMinimizer",
    "minimize_supports",
    "DefaultStrategy",
    "DefaultToyStrategy",
    "PushbackStrategy",
    "EDM",
]



#  Copyright (c) 2024 zfit
import numpy as np

import zfit


def test_to_binned_polynomial():

    low, high = 1, 10
    obs = zfit.Space('obs', low, high, binning=100)
    size_normal = 10_000
    data_normal_np = np.random.normal(size=size_normal, loc=5, scale=1)
    data_expo_np = np.random.exponential(size=size_normal, scale=20)
    data_tot_np = np.concatenate([data_normal_np, data_expo_np])
    data = zfit.Data(data_tot_np, obs=obs)

    sigma = zfit.Parameter('sigma', 1.)
    alpha = zfit.Parameter('alpha', 90., 0., 100.)
    c0 = zfit.Parameter('c0', 0.1)
    c1 = zfit.Parameter('c1', -0.026, -100., 100.)
    c2 = zfit.Parameter('c2', -0.056, -100., 100.)

    pdf_1 = zfit.pdf.GeneralizedGaussExpTail(
        obs=obs,
        mu=5.,
        sigmar=sigma,
        sigmal=sigma,
        alphar=alpha,
        alphal=alpha
    )
    pdf_2 = zfit.pdf.Chebyshev(obs=obs, coeffs=[c1, c2], coeff0=c0)
    pdf = zfit.pdf.BinnedSumPDF([pdf_1, pdf_2], [0.5], obs=obs)
    vals = pdf.pdf(data)  # make sure this doesn't crash and the spaces work in the jitted mode
    assert not np.any(np.isnan(vals.numpy()))



#  Copyright (c) 2024 zfit
import pytest
import numpy as np

import zfit
import zfit.z.numpy as znp


def test_studentt():
    N_true = 5
    mu_true = 2
    sigma_true = 1
    N = zfit.Parameter("N", N_true)
    mu = zfit.Parameter("mu", mu_true)
    sigma = zfit.Parameter("sigma", sigma_true)

    obs = zfit.Space(obs="obs", limits=(-50, 50))
    t = zfit.pdf.StudentT(ndof=N, mu=mu, sigma=sigma, obs=obs)
    test_values = np.random.uniform(low=-10.0, high=10.0, size=1000)
    samples = t.sample(100_000)["obs"]
    probs = t.pdf(x=test_values)

    assert np.all(np.isfinite(probs))
    assert np.all(probs>=0)
    assert pytest.approx(2, rel=1e-2) == znp.mean(samples)
    assert pytest.approx(np.sqrt(N_true / (N_true - 2)), rel=1e-2) == znp.std(samples)



#  Copyright (c) 2024 zfit
from __future__ import annotations

import numpy as np

import zfit

# create space
obs = zfit.Space("x", -2, 3)

# parameters
mu = zfit.Parameter("mu", 1.2, -4, 6)
sigma = zfit.Parameter("sigma", 1.3, 0.5, 10)

# model building, pdf creation
gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
obs_binned = obs.with_binning(30)
gauss_binned = gauss.to_binned(obs_binned)

normal_np = np.random.normal(loc=2.0, scale=3.0, size=10000)
data = zfit.Data.from_numpy(obs=obs, array=normal_np)
data_binned = data.to_binned(obs_binned)  # could also be a hist object

nll = zfit.loss.BinnedNLL(model=gauss_binned, data=data_binned)

minimizer = zfit.minimize.Minuit(gradient="zfit")
result = minimizer.minimize(nll).update_params()

param_errors = result.hesse()

# or here with minos
param_errors_asymmetric, new_result = result.errors()



#  Copyright (c) 2024 zfit
import pickle

import numpy as np
import pytest
import tensorflow as tf

import zfit._data.unbinneddata
import zfit._variables.axis as axis
import zfit.z.numpy as znp
from zfit.util.exception import ObsIncompatibleError


@pytest.mark.skip
def test_basic():
    n = 1000
    nobs = 3
    axes = axis.SpaceV2(
        [axis.UnbinnedAxis("x"), axis.UnbinnedAxis("y"), axis.UnbinnedAxis("z")]
    )
    tensor = znp.random.uniform(size=(n, nobs))
    weights = znp.random.uniform(size=(n,))
    data = zfit._data.unbinneddata.UnbinnedData(data=tensor, axes=axes, weights=weights)

    pickled = pickle.dumps(data)
    unpickled = pickle.loads(pickled)
    # assert data.axes == unpickled.axes
    np.testing.assert_array_equal(data.weights, unpickled.weights)
    np.testing.assert_array_equal(data.data, unpickled.data)

    array = data["y"]
    np.testing.assert_array_equal(array, tensor[:, 1])

    # af = asdf.AsdfFile(data.obj_to_repr())
    # with io.BytesIO(b'file') as f:
    #     af.write_to(f)


def test_data_conversion():
    obs1 = zfit.Space("obs1", (-1, 100))
    obs2 = zfit.Space("obs2", (-1, 100))
    obs12 = obs1 * obs2
    testdata1 = [1.0, 2.0, 3.0, 4.0, 5.0]
    from zfit.core.data import convert_to_data

    out1 = convert_to_data(testdata1, obs=obs1)
    assert out1.value().shape == (5, 1)
    np.testing.assert_allclose(out1["obs1"], testdata1)
    # check that it fails without obs
    with pytest.raises(ValueError):
        convert_to_data(testdata1)

    testdata2 = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], [1.0, 22.0, 34.0, 41.0, 5.0]]).T
    out2 = convert_to_data(testdata2, obs=obs12)
    assert out2.value().shape == (5, 2)
    np.testing.assert_allclose(out2["obs1"], testdata2[:, 0])
    np.testing.assert_allclose(out2["obs2"], testdata2[:, 1])
    with pytest.raises(ValueError):
        convert_to_data(testdata2)

    # test dataframe
    import pandas as pd

    df = pd.DataFrame(testdata2, columns=["obs1", "obs2"])
    out3 = convert_to_data(df)
    assert out3.value().shape == (5, 2)
    np.testing.assert_allclose(out3["obs1"], testdata2[:, 0])
    np.testing.assert_allclose(out3["obs2"], testdata2[:, 1])

    # test single value
    single_value = 5.0
    out4 = convert_to_data(single_value, obs=obs1)
    assert out4.value().shape == (1, 1)
    np.testing.assert_allclose(out4["obs1"], single_value)
    with pytest.raises(ValueError):
        convert_to_data(single_value)

    # test single int
    single_value = 5
    out5 = convert_to_data(single_value, obs=obs1)
    assert out5.value().shape == (1, 1)
    np.testing.assert_allclose(out5["obs1"], single_value)
    with pytest.raises(ValueError):
        convert_to_data(single_value)

def test_unbinneddata_getitem_index():
    n = 1000
    nobs = 3
    space = zfit.Space(obs="obs1", axes=(0,), limits=(-1, 1)) * zfit.Space(obs="obs2", axes=(1,), limits=(-1, 1)) * zfit.Space(obs="obs3", axes=(2,), limits=(-1, 1))
    space = space.with_autofill_axes()
    arr = znp.random.uniform(size=(n, nobs))
    weights = znp.random.uniform(size=(n,))
    data = zfit.Data.from_numpy(obs=space, array=arr, weights=weights)
    assert space.axes == (0, 1, 2)

    data0 = data.with_obs(0)
    data1 = data.with_obs(1)

    assert data0.value().shape == (n, 1)
    assert data1.value().shape == (n, 1)
    np.testing.assert_array_equal(data0.value(), arr[:, 0:1])
    np.testing.assert_array_equal(data1.value(), arr[:, 1:2])
    np.testing.assert_array_equal(data0["obs1"], arr[:, 0])
    np.testing.assert_array_equal(data1["obs2"], arr[:, 1])
    np.testing.assert_array_equal(data["obs1"], arr[:, 0])
    np.testing.assert_array_equal(data["obs2"], arr[:, 1])

    data10 = data.with_obs([1, 0])
    assert data10.value().shape == (n, 2)
    np.testing.assert_array_equal(data10.value(), arr[:, 1::-1])
    np.testing.assert_array_equal(data10["obs2"], arr[:, 1])
    np.testing.assert_array_equal(data10["obs1"], arr[:, 0])

    np.testing.assert_array_equal(data10[0], arr[:, 1])
    np.testing.assert_array_equal(data10[1], arr[:, 0])




def test_unbinned_data_concat_index():
    n = 1000
    n2 = 552
    n3 = 123
    nobs = 3
    space1 = zfit.Space(obs="obs1", limits=(-1, 1))
    space2 = zfit.Space(obs="obs2", limits=(-11, 5))
    space3 = zfit.Space(obs="obs3", limits=(-15, 3))
    space = space1 * space2 * space3

    arr1 = znp.random.uniform(size=(n, nobs))
    weights1 = znp.random.uniform(size=(n,))
    data1w = zfit.Data.from_numpy(obs=space, array=arr1, weights=weights1)
    data1now = zfit.Data.from_numpy(obs=space, array=arr1)

    arr2 = znp.random.uniform(size=(n2, nobs))
    weights2 = znp.random.uniform(size=(n2,))
    data2w = zfit.Data.from_numpy(obs=space, array=arr2, weights=weights2)
    data2now = zfit.Data.from_numpy(obs=space, array=arr2)
    # todo: if we want to switch argument orders, legacy element
    # with pytest.warns(UserWarning):
    #     data2now_order = zfit.Data.from_numpy(space, arr2)  # wrong order, legacy check
    # np.testing.assert_array_equal(data2now_order.value(), data2w.value())
    # assert data2now_order.space == data2w.space


    arr3 = znp.random.uniform(size=(n3, nobs))
    weights3 = znp.random.uniform(size=(n3,))
    data3w = zfit.Data.from_numpy(obs=space, array=arr3, weights=weights3)
    data3now = zfit.Data.from_numpy(obs=space, array=arr3)

    dataw = zfit.data.concat([data1w, data2w])
    np.testing.assert_array_equal(dataw.value(), np.concatenate([arr1, arr2]))
    assert dataw.space == space
    np.testing.assert_array_equal(dataw.weights, np.concatenate([weights1, weights2]))

    datawall = zfit.data.concat([data1w, data2w, data3w])
    np.testing.assert_array_equal(datawall.value(), np.concatenate([arr1, arr2, arr3]))
    assert datawall.space == space
    np.testing.assert_array_equal(datawall.weights, np.concatenate([weights1, weights2, weights3]))
    assert datawall.space == space
    spacemixed = space1 * space3 * space2

    mixedarray = np.concatenate([np.array(arr1)[:, [0, 2, 1]], np.array(arr2)[:, [0, 2, 1]], np.array(arr3)[:, [0, 2, 1]],], axis=0)

    datawallmixed = zfit.data.concat([data1w, data2w, data3w], obs=spacemixed)
    np.testing.assert_array_equal(datawallmixed.value(), mixedarray)
    assert datawallmixed.space == spacemixed
    np.testing.assert_array_equal(datawallmixed.weights, np.concatenate([weights1, weights2, weights3]))

    datanow = zfit.data.concat([data1now, data2now])
    np.testing.assert_array_equal(datanow.value(), np.concatenate([arr1, arr2]))
    assert datanow.space == space
    assert datanow.weights is None

    datamixedw = zfit.data.concat([data1w, data2now, data3w])
    np.testing.assert_array_equal(datamixedw.value(), np.concatenate([arr1, arr2, arr3]))
    assert datamixedw.space == space
    np.testing.assert_array_equal(datamixedw.weights, np.concatenate([weights1, np.ones_like(weights2), weights3]))

def test_unbinned_data_concat_obs():
    n = 1000
    nobs = 1
    space1 = zfit.Space(obs="obs1", limits=(-1, 1))
    space2 = zfit.Space(obs="obs2", limits=(-11, 5))
    space3 = zfit.Space(obs="obs3", limits=(-15, 3))
    space = space1 * space2 * space3

    arr1 = znp.random.uniform(size=(n, nobs))
    weights1 = znp.random.uniform(size=(n,))
    data1w = zfit.Data.from_numpy(obs=space1, array=arr1, weights=weights1)
    data1now = zfit.Data.from_numpy(obs=space1, array=arr1)

    arr2 = znp.random.uniform(size=(n, nobs))
    weights2 = znp.random.uniform(size=(n,))
    data2w = zfit.Data.from_numpy(obs=space2, array=arr2, weights=weights2)
    data2now = zfit.Data.from_numpy(obs=space2, array=arr2)

    arr3 = znp.random.uniform(size=(n, nobs))
    weights3 = znp.random.uniform(size=(n,))
    data3w = zfit.Data.from_numpy(obs=space3, array=arr3, weights=weights3)
    data3now = zfit.Data.from_numpy(obs=space3, array=arr3)

    data12w = zfit.data.concat([data1w, data2w], axis=1)
    np.testing.assert_array_equal(data12w.value(), np.concatenate([arr1, arr2], axis=1))
    assert data12w.space == space1 * space2
    np.testing.assert_array_equal(data12w.weights, weights1 * weights2)

    data123w = zfit.data.concat([data1w, data2w, data3w], axis=1)
    np.testing.assert_array_equal(data123w.value(), np.concatenate([arr1, arr2, arr3], axis=1))
    assert data123w.space == space
    np.testing.assert_array_equal(data123w.weights, weights1 * weights2 * weights3)

    spacemixed = space1 * space3 * space2
    data123mixed = zfit.data.concat([data1w, data2w, data3w], obs=spacemixed, axis=1)
    mixedarray = np.concatenate([arr1, arr3, arr2], axis=1)
    assert data123mixed.space == spacemixed
    np.testing.assert_array_equal(data123mixed.value(), mixedarray)
    np.testing.assert_allclose(data123mixed.weights, weights1 * weights3 * weights2)

    datanow = zfit.data.concat([data1now, data2now], axis=1)
    np.testing.assert_array_equal(datanow.value(), np.concatenate([arr1, arr2], axis=1))
    assert datanow.space == space1 * space2
    assert datanow.weights is None

    datamixedw = zfit.data.concat([data1w, data2now, data3w], axis=1)
    np.testing.assert_array_equal(datamixedw.value(), np.concatenate([arr1, arr2, arr3], axis=1))
    assert datamixedw.space == space
    np.testing.assert_allclose(datamixedw.weights, weights1 * weights3)

def test_unbinned_data_concat_obs_errors():

    space1 = zfit.Space(obs="obs1", limits=(-1, 1))
    space2 = zfit.Space(obs="obs2", limits=(-11, 5))
    space3 = zfit.Space(obs="obs3", limits=(-15, 3))
    space12 = space1 * space2

    arr1 = znp.random.uniform(size=(1000, 1))
    weights1 = znp.random.uniform(size=(1000,))
    data1w = zfit.Data.from_numpy(obs=space1, array=arr1, weights=weights1)

    arr2lessevents = znp.random.uniform(size=(10, 1))
    weights2 = znp.random.uniform(size=(10,))
    data2wless = zfit.Data.from_numpy(obs=space2, array=arr2lessevents, weights=weights2)

    arr2 = znp.random.uniform(size=(1000, 1))
    weights2 = znp.random.uniform(size=(1000,))
    data2w = zfit.Data.from_numpy(obs=space2, array=arr2, weights=weights2)

    arr3 = znp.random.uniform(size=(1000, 1))
    weights3 = znp.random.uniform(size=(1000,))
    data3w = zfit.Data.from_numpy(obs=space3, array=arr3, weights=weights3)



    with pytest.raises(tf.errors.InvalidArgumentError):
        zfit.data.concat([data1w, data2wless], axis=1)

    arr12 = znp.random.uniform(size=(1000, 2))
    weights12 = znp.random.uniform(size=(1000,))
    data12w = zfit.Data.from_numpy(obs=space12, array=arr12, weights=weights12)

    with pytest.raises(ObsIncompatibleError):
        zfit.data.concat([data1w, data12w], axis=1)

    with pytest.raises(ObsIncompatibleError):
        zfit.data.concat([data1w, data3w], obs=space12, axis=1)



"""Numpy like interface for math functions and arrays. This module is intended to replace tensorflow specific methods
and datastructures with equivalent or similar versions in the numpy api. This should help make zfit as a project
portable to alternatives of tensorflow should it be necessary in the future. At the moment it is simply an alias for the
numpy api of tensorflow. See https://www.tensorflow.org/guide/tf_numpy for more a guide to numpy api in tensorflow. See
https://www.tensorflow.org/api_docs/python/tf/experimental/numpy for the complete numpy api in tensorflow. Recommended
way of importing:

>>> import zfit.z.numpy as znp
"""

#  Copyright (c) 2024 zfit
from __future__ import annotations

import numpy as np
import tensorflow as tf
from tensorflow.experimental.numpy import *  # noqa: F403

# ruff: noqa: F405


class linalg:
    inv = staticmethod(tf.linalg.inv)
    det = staticmethod(tf.linalg.det)
    solve = staticmethod(tf.linalg.solve)


# TODO: move into special namespace when that's available
def faddeeva_humlicek(z, s=10.0):
    """Complex error function w(z = x + iy) combining Humlicek's rational approximations.

    |x| + y > s:  Humlicek (JQSRT, 1982) rational approximation for region II;
    else:          Humlicek (JQSRT, 1979) rational approximation with n=16 and delta=y0=1.35

    Version using tensorflow and the numpy api of tensorflow.
    Single complex argument version of Franz Schreier's cpfX.hum2zpf16m.
    Originally licensed under a 3-clause BSD style license - see
    https://atmos.eoc.dlr.de/tools/lbl4IR/cpfX.py

    Args:
        z: array_like
        s: float, optional
            The threshold for the region II approximation.
            The default is 10.0.

    Returns:
        A `Tensor` of `complex128` dtype.
    """

    z = asarray(atleast_1d(z), dtype=complex128)

    AA = array(
        [
            +46236.3358828121,
            -147726.58393079657j,
            -206562.80451354137,
            281369.1590631087j,
            +183092.74968253175,
            -184787.96830696272j,
            -66155.39578477248,
            57778.05827983565j,
            +11682.770904216826,
            -9442.402767960672j,
            -1052.8438624933142,
            814.0996198624186j,
            +45.94499030751872,
            -34.59751573708725j,
            -0.7616559377907136,
            0.5641895835476449j,
        ],
        dtype=complex128,
    )

    bb = array(
        [
            +7918.06640624997,
            -126689.0625,
            +295607.8125,
            -236486.25,
            +84459.375,
            -15015.0,
            +1365.0,
            -60.0,
            +1.0,
        ],
        dtype=complex128,
    )

    sqrt_piinv = 1.0 / np.sqrt(np.pi)

    zz = z * z
    w = 1j * z * (zz * sqrt_piinv - 1.410474) / (0.75 + zz * (zz - 3.0))

    real_part = real(z)
    imag_part = imag(z)
    mask = logical_and(abs(real_part) + imag_part < s, imag_part < s)

    Z = z + 1.35j
    ZZ = Z * Z

    numer = (
        (
            (
                (
                    (
                        (
                            (
                                (
                                    (
                                        (
                                            ((((AA[15] * Z + AA[14]) * Z + AA[13]) * Z + AA[12]) * Z + AA[11]) * Z
                                            + AA[10]
                                        )
                                        * Z
                                        + AA[9]
                                    )
                                    * Z
                                    + AA[8]
                                )
                                * Z
                                + AA[7]
                            )
                            * Z
                            + AA[6]
                        )
                        * Z
                        + AA[5]
                    )
                    * Z
                    + AA[4]
                )
                * Z
                + AA[3]
            )
            * Z
            + AA[2]
        )
        * Z
        + AA[1]
    ) * Z + AA[0]

    denom = (
        ((((((ZZ + bb[7]) * ZZ + bb[6]) * ZZ + bb[5]) * ZZ + bb[4]) * ZZ + bb[3]) * ZZ + bb[2]) * ZZ + bb[1]
    ) * ZZ + bb[0]

    return where(mask, numer / denom, w)



#  Copyright (c) 2024 zfit

from __future__ import annotations

from zfit.util.temporary import TemporarilySet


class JIT:
    def _set_all(self, enable: bool = True):
        new_values = {k: enable for k in self._get_allowed()}

        def getter():
            return self._get_allowed().copy()

        def setter(jit_types):
            self._update_allowed(jit_types)

        return TemporarilySet(getter=getter, setter=setter, value=new_values)

    def _set_default(self):
        from zfit import z

        new_values = z.zextension.FunctionWrapperRegistry._DEFAULT_DO_JIT_TYPES.copy()

        for key in self._get_allowed():
            if key not in new_values:
                new_values[key] = new_values[key]  # default dict will explicitly set the default value

        def getter():
            return self._get_allowed().copy()

        def setter(jit_types):
            self._update_allowed(jit_types)

        return TemporarilySet(getter=getter, setter=setter, value=new_values)

    def _update_allowed(self, update_jit):
        from zfit import z

        z.zextension.FunctionWrapperRegistry.do_jit_types.update(update_jit)

    def _get_allowed(self):
        from zfit import z

        return z.zextension.FunctionWrapperRegistry.do_jit_types

    @property
    def experimental_is_eager(self):
        from ..settings import run

        return run.mode["graph"]


jit = JIT()  # singleton



#  Copyright (c) 2024 zfit
from __future__ import annotations

from .minimizers.errors import compute_errors
from .minimizers.fitresult import Approximations, FitResult

__all__ = ["FitResult", "compute_errors", "Approximations"]



#  Copyright (c) 2024 zfit
import numpy as np
import pytest
import scipy.stats

import zfit
from zfit import z
from zfit.core.constraint import BaseConstraint, GaussianConstraint, SimpleConstraint, _preprocess_gaussian_constr_sigma_var
import zfit.z.numpy as znp
from zfit.core.constraint import BaseConstraint, GaussianConstraint, SimpleConstraint
from zfit.util.container import convert_to_container
from zfit.util.exception import ShapeIncompatibleError


def true_nll_gaussian(x, mu, sigma):
    x = convert_to_container(x, container=tuple)
    mu = convert_to_container(mu, container=tuple)
    sigma = convert_to_container(sigma, container=tuple)
    constraint = z.constant(0.0)
    if not len(x) == len(mu) == len(sigma):
        raise ValueError("params, mu and sigma have to have the same length.")
    for x_, mean, sig in zip(x, mu, sigma):
        constraint += z.reduce_sum(z.square(x_ - mean) / (2.0 * z.square(sig)))

    return constraint


def true_gauss_constr_value(x, mu, sigma):
    logpdf = lambda x, loc, scale: scipy.stats.norm.logpdf(x, loc=loc, scale=scale)
    return -np.sum(
        [logpdf(x_, loc=mu, scale=sigma) for x_, mu, sigma in zip(x, mu, sigma)]
    )


def true_poisson_constr_value(x, lam):
    logpdf = lambda x, loc: scipy.stats.poisson.logpmf(x, mu=loc)
    return -np.sum([logpdf(x_, loc=lam_) for x_, lam_ in zip(x, lam)])


def true_multinormal_constr_value(x, mean, cov):
    return -scipy.stats.multivariate_normal.logpdf(x, mean=mean, cov=cov)


def test_base_constraint():
    with pytest.raises(TypeError):
        BaseConstraint()


def test_poisson_constrain():
    x, lam = np.random.randint(1, 100, size=(2, 50))
    constr = zfit.constraint.PoissonConstraint(
        params=z.convert_to_tensor(x), observation=z.convert_to_tensor(lam)
    )
    poiss_constr_val = constr.value()
    true_val = true_poisson_constr_value(x, lam)
    np.testing.assert_allclose(poiss_constr_val, true_val)


def test_gaussian_constraint_shape_errors():
    param1 = zfit.Parameter("Param1", 5)
    param2 = zfit.Parameter("Param2", 6)

    obs1 = zfit.Parameter("obs1", 2)
    obs2 = zfit.Parameter("obs2", 3)
    obs3 = zfit.Parameter("obs3", 4)

    with pytest.raises(ShapeIncompatibleError):
        GaussianConstraint(
            params=[param1, param2], observation=[obs1, obs2, obs3], uncertainty=5
        ).value()
    with pytest.raises(ShapeIncompatibleError):
        GaussianConstraint(
            params=[param1, param2], observation=[obs1, obs3], uncertainty=5
        ).value()
    with pytest.raises(ShapeIncompatibleError):
        GaussianConstraint(
            params=[param1, param2], observation=obs1, uncertainty=[1, 4]
        ).value()
    with pytest.raises(ShapeIncompatibleError):
        GaussianConstraint(
            params=param1, observation=[obs1, obs3], uncertainty=[2, 3]
        ).value()


def test_gaussian_constraint_matrix_legacy():
    param1 = zfit.Parameter("Param1", 5)
    param2 = zfit.Parameter("Param2", 6)
    params = [param1, param2]

    observed = [3.0, 6.1]
    sigma = np.array([[1, 0.3], [0.3, 0.5]])

    trueval = true_multinormal_constr_value(
        x=znp.asarray(params), mean=observed, cov=sigma
    )

    constr = GaussianConstraint(params=params, observation=observed, uncertainty=sigma)
    constr_np = znp.asarray(constr.value())
    assert pytest.approx(trueval) == constr_np
    # assert constr_np == pytest.approx(3.989638)

    assert set(constr.get_params()) == set(params)


@pytest.mark.parametrize("kwargs", [{'sigma': np.array([[1, 0.3], [0.3, 0.5]]) ** 0.5}, {'cov': np.array([[1, 0.3], [0.3, 0.5]])}], ids=['sigma', 'cov'])
def test_gaussian_constraint_matrix(kwargs):
    param1 = zfit.Parameter("Param1", 5)
    param2 = zfit.Parameter("Param2", 6)
    params = [param1, param2]

    observed = [3.0, 6.1]
    sigma = np.array([[1, 0.3], [0.3, 0.5]])

    trueval = true_multinormal_constr_value(
        x=zfit.run(params), mean=observed, cov=sigma
    )
    if 'sigma' in kwargs:
        with pytest.raises(ValueError):
            _ = GaussianConstraint(params=params, observation=observed, **kwargs)
        return
    constr = GaussianConstraint(params=params, observation=observed, **kwargs)
    constr_np = zfit.run(constr.value())
    assert constr_np == pytest.approx(trueval)
    # assert constr_np == pytest.approx(3.989638)

    assert set(constr.get_params()) == set(params)



def test_gausian_constraint_matrix_preprocess():

    covin = np.array([[1, 0.3], [0.3, 0.5]])
    sigma, cov = _preprocess_gaussian_constr_sigma_var(covin, None, False)
    np.testing.assert_allclose(sigma, np.sqrt([covin[0, 0], covin[1, 1]]))
    np.testing.assert_allclose(cov, covin)

    covin = np.array([2.0, 0.3])
    sigma, cov = _preprocess_gaussian_constr_sigma_var(covin, None, False)
    np.testing.assert_allclose(sigma ** 2, covin)
    np.testing.assert_allclose(cov, np.array([[covin[0], 0], [0, covin[1]]]))

    covin = np.array([9.])
    sigma, cov = _preprocess_gaussian_constr_sigma_var(covin, None, False)
    np.testing.assert_allclose(sigma ** 2, covin)
    np.testing.assert_allclose(cov, np.array([[covin[0]]]))

    covin = 8.
    sigma, cov = _preprocess_gaussian_constr_sigma_var(covin, None, False)
    np.testing.assert_allclose(sigma ** 2, np.array([covin]))
    np.testing.assert_allclose(cov, np.array([[covin]]))

    covin = [1, 0.3]
    sigma, cov = _preprocess_gaussian_constr_sigma_var(covin, None, False)
    np.testing.assert_allclose(sigma, np.sqrt(covin))
    np.testing.assert_allclose(cov, np.diag(covin))

    sigmain = np.array([1, 0.3])
    sigma, cov = _preprocess_gaussian_constr_sigma_var(None, sigmain, False)
    np.testing.assert_allclose(sigma, sigmain)
    np.testing.assert_allclose(cov, np.array([[sigmain[0], 0], [0, sigmain[1]]]) ** 2)

    sigmain = 3.
    sigma, cov = _preprocess_gaussian_constr_sigma_var(None, sigmain, False)
    np.testing.assert_allclose(sigma, np.array([sigmain]))
    np.testing.assert_allclose(cov, np.array([[sigmain]]) ** 2)





def test_gaussian_constraint_legacy():
    param_vals = [5, 6, 3]
    observed = [zfit.Parameter(f"observed {val}", val) for val in [3, 6.1, 4.3]]
    sigma = [1, 0.3, 0.7]
    true_val = true_gauss_constr_value(x=observed, mu=param_vals, sigma=sigma)
    assert true_val == true_gauss_constr_value(x=param_vals, mu=observed, sigma=sigma)
    params = [zfit.Parameter(f"Param{i}", val) for i, val in enumerate(param_vals)]

    constr = GaussianConstraint(params=params, observation=observed, uncertainty=sigma)
    constr_np = constr.value()
    assert pytest.approx(true_val) == constr_np
    assert set(constr.get_params()) == set(params)

    param_vals[0] = 2
    params[0].set_value(param_vals[0])

    constr2_np = constr.value()
    constr2_newtensor_np = constr.value()
    assert pytest.approx(constr2_np) == constr2_newtensor_np

    true_val2 = true_gauss_constr_value(x=param_vals, mu=observed, sigma=sigma)
    assert pytest.approx(true_val2) == constr2_np

    constr.observation[0].set_value(5)
    observed[0] = 5
    # print("x: ", param_vals, [p for p in params])
    # print("mu: ", observed, [p for p in constr.observation])
    # print("sigma: ", sigma, np.sqrt([p for p in np.diag(constr.covariance)]))
    true_val3 = true_gauss_constr_value(x=param_vals, mu=observed, sigma=sigma)
    constr3_np = constr.value()
    assert pytest.approx(true_val3) == constr3_np

@pytest.mark.parametrize("kwargs", [{'sigma': [1, 0.3, 0.7]}, {'cov': np.array([1, 0.3, 0.7]) ** 2},
                                    {'cov': np.array([[1, 0, 0], [0, 0.3, 0], [0, 0, 0.7]]) ** 2}],
                         ids=['sigma', 'cov', 'cov_matrix'])
def test_gaussian_constraint(kwargs):
    param_vals = [5, 6, 3]
    observed = [zfit.Parameter(f"observed {val}", val) for val in [3, 6.1, 4.3]]
    sigma = [1, 0.3, 0.7]
    true_val = true_gauss_constr_value(x=observed, mu=param_vals, sigma=sigma)
    assert true_val == true_gauss_constr_value(x=param_vals, mu=observed, sigma=sigma)
    params = [zfit.Parameter(f"Param{i}", val) for i, val in enumerate(param_vals)]

    constr = GaussianConstraint(params=params, observation=observed, **kwargs)
    constr_np = constr.value().numpy()
    assert constr_np == pytest.approx(true_val)
    assert set(constr.get_params()) == set(params)

    param_vals[0] = 2
    params[0].set_value(param_vals[0])

    constr2_np = constr.value().numpy()
    constr2_newtensor_np = constr.value().numpy()
    assert constr2_newtensor_np == pytest.approx(constr2_np)

    true_val2 = true_gauss_constr_value(x=param_vals, mu=observed, sigma=sigma)
    assert constr2_np == pytest.approx(true_val2)

    constr.observation[0].set_value(5)
    observed[0] = 5
    true_val3 = true_gauss_constr_value(x=param_vals, mu=observed, sigma=sigma)
    constr3_np = constr.value().numpy()
    assert constr3_np == pytest.approx(true_val3)


sigma_true_orderbug = [0.05 * 1500, 0.001, 0.01, 0.1, 0.05 * 0.5]


@pytest.mark.parametrize("kwargs", [{'sigma': sigma_true_orderbug},
                                    {'cov': np.diag(sigma_true_orderbug) ** 2}],
                            ids=['sigma', 'cov'])
def test_gaussian_constraint_orderbug(kwargs):  # as raised in #162
    observed = [1500, 1.0, 1.0, 1.0, 0.5]
    params = [zfit.Parameter(f"param{i}", val) for i, val in enumerate(observed)]

    sigma = sigma_true_orderbug
    true_val = true_gauss_constr_value(x=observed, mu=observed, sigma=sigma)

    constr1 = GaussianConstraint(params=params, observation=observed, **kwargs)

    value_tensor = constr1.value()
    constr_np = value_tensor.numpy()
    assert constr_np == pytest.approx(true_val)
    assert true_val < 10000

def test_gaussian_constraint_orderbug_legacy():  # as raised in #162
    observed = [1500, 1.0, 1.0, 1.0, 0.5]
    params = [zfit.Parameter(f"param{i}", val) for i, val in enumerate(observed)]

    sigma = sigma_true_orderbug
    true_val = true_gauss_constr_value(x=observed, mu=observed, sigma=sigma)

    constr1 = GaussianConstraint(params=params, observation=observed, uncertainty=sigma)

    assert pytest.approx(true_val) == constr1.value()
    assert true_val < 10000


def test_gaussian_constraint_orderbug2_legacy():  # as raised in #162, failed before fixing
    param1 = zfit.Parameter("param1", 1500)
    param5 = zfit.Parameter("param2", 0.5)

    param2 = zfit.Parameter("param3", 1.0)
    param3 = zfit.Parameter("param4", 1.0)
    param4 = zfit.Parameter("param5", 1.0)

    constraint = {
        "params": [param1, param2, param3, param4, param5],
        "observation": [1500, 1.0, 1.0, 1.0, 0.5],
        "uncertainty": sigma_true_orderbug,
    }

    constr1 = GaussianConstraint(**constraint)
    # param_vals = [1500, 1.0, 1.0, 1.0, 0.5]
    constraint["x"] = constraint["params"]

    true_val = true_gauss_constr_value(
        x=constraint["x"], mu=constraint["observation"], sigma=constraint["uncertainty"]
    )

    assert pytest.approx(true_val) == constr1.value()
    assert true_val < 1000
    assert pytest.approx(
        -8.592, abs=0.1
    ) == true_val  # if failing, change value. Hardcoded for additional layer


@pytest.mark.parametrize("kwargs", [{'sigma': sigma_true_orderbug},
                                    {'cov': np.diag(sigma_true_orderbug) ** 2},
                                    {'cov': np.array(sigma_true_orderbug) ** 2}],                            ids=['sigma', 'cov_matrix', 'cov'])
def test_gaussian_constraint_orderbug2(kwargs):  # as raised in #162, failed before fixing
    param1 = zfit.Parameter("param1", 1500)
    param5 = zfit.Parameter("param2", 0.5)

    param2 = zfit.Parameter("param3", 1.0)
    param3 = zfit.Parameter("param4", 1.0)
    param4 = zfit.Parameter("param5", 1.0)

    constraint = {
        "params": [param1, param2, param3, param4, param5],
        "observation": [1500, 1.0, 1.0, 1.0, 0.5],
        **kwargs
    }

    constr1 = GaussianConstraint(**constraint)
    # param_vals = [1500, 1.0, 1.0, 1.0, 0.5]
    constraint["x"] = [m.numpy() for m in constraint["params"]]

    true_val = true_gauss_constr_value(
        x=constraint["x"], mu=constraint["observation"], sigma=sigma_true_orderbug
    )

    value_tensor = constr1.value()
    constr_np = value_tensor.numpy()
    assert constr_np == pytest.approx(true_val)
    assert true_val < 1000
    assert true_val == pytest.approx(
        -8.592, abs=0.1
    )  # if failing, change value. Hardcoded for additional layer


@pytest.mark.flaky(3)
@pytest.mark.parametrize("kwargs", [{'sigma': [1, 0.3, 0.7]}, {'cov': np.array([1, 0.3, 0.7]) ** 2},
                                    {'cov': np.array([[1, 0, 0], [0, 0.3, 0], [0, 0, 0.7]]) ** 2}],                  ids=['sigma', 'cov', 'cov_matrix'])
def test_gaussian_constraint_sampling(kwargs):
    param1 = zfit.Parameter("Param1", 5)
    params = [param1]

    observed = [5]
    sigma = [1]
    constr = GaussianConstraint(params=params, observation=observed, **kwargs)

    sample = constr.sample(15000)

    assert np.mean(sample[param1]) == pytest.approx(observed[0], rel=0.01)
    assert np.std(sample[param1]) == pytest.approx(sigma[0], rel=0.01)

@pytest.mark.flaky(3)
def test_gaussian_constraint_sampling_legacy():
    param1 = zfit.Parameter("Param1", 5)
    params = [param1]

    observed = [5]
    sigma = [1]
    constr = GaussianConstraint(params=params, observation=observed, uncertainty=sigma)

    sample = constr.sample(15000)

    assert pytest.approx(observed[0], rel=0.01) == np.mean(sample[param1])
    assert pytest.approx(sigma[0], rel=0.01) == np.std(sample[param1])


def test_simple_constraint_legacy():
    param1 = zfit.Parameter("Param1", 5)
    param2 = zfit.Parameter("Param2", 6)
    params = [param1, param2]

    observed = [3.0, 6.1]
    sigma = [1.0, 0.5]

    def func():
        return true_nll_gaussian(x=observed, mu=params, sigma=sigma)

    constr = SimpleConstraint(func=func, params=params)

    constr_np = constr.value()
    assert pytest.approx(2.02) == constr_np

    assert set(constr.get_params()) == set(params)

def test_gauss_fails_params():
    param1 = zfit.Parameter("Param1", 5)
    param2 = zfit.Parameter("Param2", 6)
    params = [param1, param2]
    sigma = [1.0, param1]
    observed = [3.0, 6.1]
    with pytest.raises(ValueError):
        GaussianConstraint(params=params, observation=observed, sigma=sigma)
    cov = [2.0, param2]
    with pytest.raises(ValueError):
        GaussianConstraint(params=params, observation=observed, cov=cov)


def test_simple_constraint_paramfunc():
    param1 = zfit.Parameter("Param1", 5)
    param2 = zfit.Parameter("Param2", 6)
    params = {"p1": param1, "p2": param2}

    observed = [3.0, 6.1]
    sigma = [1.0, 0.5]

    def func(params):
        return true_nll_gaussian(
            x=observed, mu=[params["p1"], params["p2"]], sigma=sigma
        )

    constr = SimpleConstraint(func=func, params=params)

    assert pytest.approx(2.02) == constr.value()

    assert set(constr.get_params()) == set(params.values())


def test_log_normal_constraint():
    # x, lam = np.random.randint(1, 100, size=(2, 50))
    lam = 44.3
    x = 45.3
    lam_tensor = z.convert_to_tensor(lam)
    constr = zfit.constraint.LogNormalConstraint(
        params=z.convert_to_tensor(x),
        observation=lam_tensor,
        uncertainty=lam_tensor**0.5,
    )
    lognormal_constr_val = constr.value()
    # true_val = true_poisson_constr_value(x, lam)
    true_lognormal = 25.128554  # maybe add dynamically?
    np.testing.assert_allclose(lognormal_constr_val, true_lognormal)



#  Copyright (c) 2024 zfit
import numpy as np


def test_dumpload(tmp_path):
    import zfit
    tmpfile = tmp_path / "test_dill_dumpload1.dill"

    def create_and_fit_data(_mean, _sigma, label="", extended=True):
        """This is a dummy function to create some dataset, fit it and return the fitterd pdf and the result of the fit.

        Takes as argument the mean and sigma of the dummy dataset
        """

        obs = zfit.Space("x", -2, 3)
        mu1 = zfit.Parameter("mu1" + label, 1.2, -4, 6)
        sigma1 = zfit.Parameter("sigma1" + label, 1.3, 0.5, 10)
        yield1 = zfit.Parameter("yield1" + label, 1000, 0, 10000)

        gauss1 = zfit.pdf.Gauss(mu=mu1, sigma=sigma1, obs=obs, extended=yield1)

        mu2 = zfit.Parameter("mu2" + label, 1.2, -4, 6, floating=False)
        sigma2 = zfit.Parameter("sigma2" + label, 1.3, 0.5, 10, floating=False)
        yield2 = zfit.Parameter("yield2" + label, 1000, 0, 10000, floating=False)
        gauss2 = zfit.pdf.Gauss(mu=mu2, sigma=sigma2, obs=obs, extended=yield2)

        mu3 = zfit.Parameter("mu3" + label, 1.2, -4, 6, floating=False)
        sigma3 = zfit.Parameter("sigma3" + label, 1.3, 0.5, 10, floating=False)
        yield3 = zfit.Parameter("yield3" + label, 1000, 0, 10000, floating=False)
        gauss3 = zfit.pdf.Gauss(mu=mu3, sigma=sigma3, obs=obs, extended=yield3)

        mu4 = zfit.Parameter("mu4" + label, 1.2, -4, 6, floating=False)
        sigma4 = zfit.Parameter("sigma4" + label, 1.3, 0.5, 10, floating=False)
        yield4 = zfit.Parameter("yield4" + label, 1000, 0, 10000, floating=False)
        gauss4 = zfit.pdf.Gauss(mu=mu4, sigma=sigma4, obs=obs, extended=yield4)

        mu5 = zfit.Parameter("mu5" + label, 1.2, -4, 6, floating=False)
        sigma5 = zfit.Parameter("sigma5" + label, 1.3, 0.5, 10, floating=False)
        yield5 = zfit.Parameter("yield5" + label, 1000, 0, 10000, floating=False)
        gauss5 = zfit.pdf.Gauss(mu=mu5, sigma=sigma5, obs=obs, extended=yield5)

        frac1 = zfit.Parameter("frac1" + label, 0.5, 0, 1)
        frac2 = zfit.Parameter("frac2" + label, 0.5, 0, 1)
        frac3 = zfit.Parameter("frac3" + label, 0.5, 0, 1)

        if extended:
            model1 = zfit.pdf.SumPDF([gauss1, gauss2])
            model2 = zfit.pdf.SumPDF([gauss3, gauss4, gauss5])
            gauss = zfit.pdf.SumPDF([model1, model2])
        else:
            model1 = zfit.pdf.SumPDF([gauss1, gauss2], fracs=frac1)
            model2 = zfit.pdf.SumPDF([gauss3, gauss4, gauss5], fracs=[frac2, frac3])
            gauss = zfit.pdf.SumPDF([model1, model2], fracs=0.9)

        data = np.random.normal(loc=_mean, scale=_sigma, size=1000)

        data = obs.filter(data)

        # create NLL
        if extended:
            nll = zfit.loss.ExtendedUnbinnedNLL(model=gauss, data=data)
        else:
            nll = zfit.loss.UnbinnedNLL(model=gauss1, data=data)

        # create a minimizer
        minimizer = zfit.minimize.Minuit(gradient=False)
        result = minimizer.minimize(nll).update_params()
        for p in nll.get_params(floating=None):
            p.floating = True

        return result, gauss

    for ntry in range(3):
        results_dict = {}
        result1, gauss1 = create_and_fit_data(2.0, 3.0, label="1")

        results_dict["fit_result1"] = result1
        results_dict["model1"] = gauss1

        with open(tmpfile, "wb") as f:
            zfit.dill.dump(results_dict, f, verify=False)

        with open(tmpfile, "rb") as f:
            results_dict_again = zfit.dill.load(f)
        result2, gauss2 = create_and_fit_data(3.0, 2.0, label="2")
        results_dict_again["fit_result2"] = result2
        results_dict_again["model2"] = gauss2

        # this is the test, this should not fail
        out = zfit.dill.dumps(results_dict_again, verify=True)
        loaded = zfit.dill.loads(out)
        assert loaded is not None
        with open(tmpfile, "w+b") as f:
            zfit.dill.dump(results_dict_again, f, verify=True)

        with open(tmpfile, "rb") as f:
            loaded = zfit.dill.load(f)

        assert loaded is not None



#  Copyright (c) 2024 zfit

import pytest

import numpy as np

import zfit
import zfit.z.numpy as znp


@pytest.fixture
def pdf():
    from zfit.pdf import Gauss

    mu = zfit.Parameter("mu", 0)
    sigma = zfit.Parameter("sigma", 1)
    return Gauss(mu=mu, sigma=sigma, obs=zfit.Space("obs1", (-10, 10)))


@pytest.fixture
def extpdf():
    from zfit.pdf import Gauss

    mu = zfit.Parameter("mu", 0)
    sigma = zfit.Parameter("sigma", 1)
    return Gauss(mu=mu, sigma=sigma, obs=zfit.Space("obs1", (-10, 10)), extended=zfit.Parameter("extended", 1004))


@pytest.fixture
def limits() -> list[zfit.Space]:
    return [zfit.Space("obs1", (-1, 1)), zfit.Space("obs1", (2, 3))]


def test_truncated_pdf_initialization(pdf, limits):
    truncated_pdf = zfit.pdf.TruncatedPDF(pdf, limits=limits)
    assert truncated_pdf.pdfs[0] == pdf
    assert truncated_pdf._limits == tuple(limits)


def test_truncated_pdf_unnormalized_pdf(pdf, limits):
    truncated_pdf = zfit.pdf.TruncatedPDF(pdf, limits)
    x = np.array([-0.5, 0.5, 2.5])
    prob = truncated_pdf.pdf(x)
    assert np.all(prob >= 0)
    # test all outside limits
    x = np.array([-2, 1.5, 4])
    prob = truncated_pdf.pdf(x)
    assert np.all(prob == 0)


def test_truncated_pdf_normalization(pdf, limits):
    truncated_pdf = zfit.pdf.TruncatedPDF(pdf, limits)
    norm = truncated_pdf.normalization(limits[0])
    assert np.all(norm >= 0)


@pytest.mark.parametrize("wrapper", [True, False], ids=["wrapper", "no_wrapper"])
def test_truncated_pdf_normalization_ext(extpdf, pdf, limits, wrapper):
    if wrapper:
        truncated_pdf = zfit.pdf.TruncatedPDF(extpdf, limits=limits)
    else:
        truncated_pdf = extpdf.to_truncated(limits=limits)
    x = znp.linspace(*limits[0].v1.limits, num=100)
    prob = truncated_pdf.ext_pdf(x)
    prob_true = extpdf.ext_pdf(x)
    np.testing.assert_allclose(prob, prob_true)
    assert truncated_pdf.extended
    assert pytest.approx(truncated_pdf.ext_integrate(limits[0])) == extpdf.ext_integrate(limits[0])
    norm = truncated_pdf.normalization(limits[0])
    assert np.all(norm >= 0)

@pytest.mark.parametrize("limits", [None, True], ids=["limits_none", "equivalent_limits"])
def test_pdf_to_truncated(extpdf, limits):
    if limits:
        lim = [extpdf.space.v1.lower, -2, 4, 4.5, extpdf.space.v1.upper]

        limits = [zfit.Space("obs1", lower=lim[i], upper=lim[i + 1]) for i in range(0, len(lim) - 1)]
    truncated_pdf = extpdf.to_truncated(limits=limits)
    space = extpdf.space
    assert truncated_pdf.integrate(space) == extpdf.integrate(space)
    x = np.array([-0.5, 0.5, 2.5])
    prob = truncated_pdf.pdf(x)
    assert np.all(prob >= 0)
    # test all outside limits
    x = np.array([space.v1.lower -5, space.v1.upper +3])
    prob = truncated_pdf.pdf(x)
    assert np.all(prob == 0)
    sample = truncated_pdf.sample(n=1000000)
    sample_true = extpdf.sample(n=1000000)
    assert pytest.approx(np.mean(sample.value()), abs=0.01) == np.mean(sample_true.value())
    assert pytest.approx(np.std(sample.value()), abs=0.01) == np.std(sample_true.value())



def test_truncated_pdf_outside_limits(pdf, limits):
    truncated_pdf = zfit.pdf.TruncatedPDF(pdf, limits)
    x = np.array([-2, 1.5, 4])
    prob = truncated_pdf.pdf(x)
    assert np.all(prob == 0)


def test_truncated_pdf_integrate(pdf, limits):
    truncated_pdf = zfit.pdf.TruncatedPDF(pdf, limits)
    integral = truncated_pdf.integrate(limits[0])
    assert np.all(integral >= 0)


def test_truncated_pdf_sample(pdf, limits):
    truncated_pdf = zfit.pdf.TruncatedPDF(pdf, limits)
    space = pdf.space
    n = 10_000
    samples = truncated_pdf.sample(n, limits[0]).value()
    assert np.all(samples >= -1)
    assert np.all(samples <= 1)

    samples = truncated_pdf.sample(n, limits[1]).value()
    assert np.all(samples >= 2)
    assert np.all(samples <= 3)

    samples = truncated_pdf.sample(n, space)[pdf.obs[0]]
    samples = np.array(samples)
    all_inside = (samples >= -1) * (samples <= 1) + (samples >= 2) * (samples <= 3)
    np.testing.assert_array_equal(all_inside, np.ones(n, dtype=bool))

    space2 = zfit.Space("obs1", (-5, 3))
    samples = np.array(truncated_pdf.sample(n, space2)[pdf.obs[0]])
    all_inside = (samples >= -1) * (samples <= 1) + (samples >= 2) * (samples <= 3)
    np.testing.assert_array_equal(all_inside, np.ones(n, dtype=bool))

def test_dynamic_truncated_yield():
    import numpy as np

    import zfit

    obs = zfit.Space("x", -10, 10)

    mean = 3
    obs1 = zfit.Space("x", -10, mean)
    obs2 = zfit.Space("x", mean, 10)


    # parameters
    mu_shared = zfit.Parameter("mu_shared", 2.0, -4, 6)  # mu is a shared parameter
    sigma1 = zfit.Parameter("sigma_one", 1.0, 0.1, 10)
    globyield = zfit.Parameter("yield1", 1100, 500, 2000)


    # model building, pdf creation
    gauss1 = zfit.pdf.Gauss(mu=mu_shared, sigma=sigma1, obs=obs, extended=globyield)
    gauss2 = zfit.pdf.Gauss(mu=mu_shared, sigma=sigma1, obs=obs, extended=globyield)

    # data
    normal_np = np.random.normal(loc=2.0, scale=3.0, size=1000)
    data = zfit.Data.from_numpy(obs=obs, array=normal_np)
    data1 = zfit.Data.from_numpy(obs=obs1, array=normal_np)  # data for the first model, can also be just a numpy array
    # the data objects just makes sure that the data is within the limits

    mean = 5
    gauss1_trunc = gauss1.to_truncated(limits=(-10, mean))
    gauss2_trunc = gauss1.to_truncated(limits=(mean, 10))

    # data
    data2 = zfit.Data.from_numpy(obs=obs2, array=normal_np)


    nll1 = zfit.loss.ExtendedUnbinnedNLL(model=gauss1_trunc, data=data1)
    nll2 = zfit.loss.ExtendedUnbinnedNLL(model=gauss2_trunc, data=data2)
    nll_simultaneous2 = nll1 + nll2

    minimizer = zfit.minimize.Minuit(tol=1e-4, gradient='zfit')
    result = minimizer.minimize(nll_simultaneous2)
    result.hesse()

    ntrue = data1.nevents + data2.nevents
    assert pytest.approx(result.params[globyield]['value'], abs=0.3) == ntrue

    nll_norm = zfit.loss.ExtendedUnbinnedNLL(model=gauss1, data=data)
    result = minimizer.minimize(nll_norm)
    result.hesse()
    assert pytest.approx(result.params[globyield]['value'], abs=0.3) == ntrue



#  Copyright (c) 2024 zfit
from __future__ import annotations

import zfit.z.numpy as znp

from ..core.interfaces import ZfitBinnedPDF
from ..core.space import supports
from ..util import ztyping
from ..util.exception import SpecificFunctionNotImplemented
from ..util.ztyping import ExtendedInputType, NormInputType
from ..z.interpolate_spline import interpolate_spline
from .functor import BaseFunctor


class SplinePDF(BaseFunctor):
    def __init__(
        self,
        pdf: ZfitBinnedPDF,
        order: int | None = None,
        obs: ztyping.ObsTypeInput = None,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str | None = "SplinePDF",
        label: str | None = None,
    ) -> None:
        """Spline interpolate a binned PDF in order to get a smooth, unbinned PDF.

        Args:
            pdf: Binned PDF that will be interpolated.
            order: Spline interpolation order. Default is 3
            obs: Unbinned observable. If not given, the observable of the pdf is used without the binning.
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| The normalization of the PDF. If this is parameter-like, it will be used as the
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        if extended is None:
            extended = pdf.is_extended
        if extended is True:
            extended = pdf.get_yield()
            self._automatically_extended = True
        else:
            self._automatically_extended = False
        if obs is None:
            obs = pdf.space
            obs = obs.with_binning(None)
        if label is None:
            label = f"splined_{pdf.label}"
        super().__init__(pdfs=pdf, obs=obs, extended=extended, norm=norm, name=name, label=label)
        if order is None:
            order = 3
        self._order = order

    @property
    def order(self):
        return self._order

    @supports(norm=True)
    def _ext_pdf(self, x, norm):
        if not self._automatically_extended:
            raise SpecificFunctionNotImplemented
        pdf = self.pdfs[0]
        density = pdf.ext_pdf(x.space, norm=norm)
        density_flat = znp.reshape(density, (-1,))
        centers_list = znp.meshgrid(*pdf.space.binning.centers, indexing="ij")
        centers_list_flat = [znp.reshape(cent, (-1,)) for cent in centers_list]
        centers = znp.stack(centers_list_flat, axis=-1)
        # [None, :, None]  # TODO: only 1 dim now
        probs = interpolate_spline(
            train_points=centers[None, ...],
            train_values=density_flat[None, :, None],
            query_points=x.value()[None, ...],
            order=self.order,
        )
        return probs[0, ..., 0]

    @supports(norm=True)
    def _pdf(self, x, norm):
        pdf = self.pdfs[0]
        density = pdf.pdf(x.space, norm=norm)  # TODO: order? Give obs, pdf makes order and binning herself?
        centers = pdf.space.binning.centers[0][None, :, None]  # TODO: only 1 dim now
        probs = interpolate_spline(
            train_points=centers,
            train_values=density[None, :, None],
            query_points=x.value()[None, ...],
            order=3,
        )
        return probs[0, ..., 0]



#  Copyright (c) 2024 zfit
import pytest
import numpy as np
from scipy.special import gamma as gamma_scipy

import zfit
import zfit.z.numpy as znp


def test_generalizedgauss():
    mu_true = 2
    sigma_true = 1
    beta_true = 2.5
    mu = zfit.Parameter("mu", mu_true)
    sigma = zfit.Parameter("sigma", sigma_true)
    beta = zfit.Parameter("beta", beta_true)

    obs = zfit.Space(obs="obs", limits=(-15, 15))
    t = zfit.pdf.GeneralizedGauss(mu=mu, sigma=sigma, beta=beta, obs=obs)
    test_values = np.random.uniform(low=-15.0, high=15.0, size=1000)
    samples = t.sample(100_000)["obs"]
    probs = t.pdf(x=test_values)

    assert np.all(np.isfinite(probs))
    assert np.all(probs>=0)
    assert pytest.approx(2, rel=1e-2) == znp.mean(samples)
    assert pytest.approx(sigma * np.sqrt(gamma_scipy(3/2.5) / gamma_scipy(1/2.5)), rel=1e-2) == znp.std(samples)



#  Copyright (c) 2024 zfit
import pytest

import zfit
import zfit.z.numpy as znp


def test_chisquared():

    N_true = 5
    N = zfit.Parameter("N", N_true)

    obs = zfit.Space(obs="obs", limits=(0, 100))
    chi2 = zfit.pdf.ChiSquared(obs=obs, ndof=N)

    samples = chi2.sample(10000)["obs"]

    assert pytest.approx(N_true, rel=0.05) == znp.mean(samples)
    assert pytest.approx((2 * N_true) ** 0.5, rel=0.05) == znp.std(samples)



#  Copyright (c) 2024 zfit
import pytest
from matplotlib import pyplot as plt

import zfit
from zfit.util import plotter

# todo: add actual assertions?

folder = "plotting"
def test_plot_1d_simple():
    obs = zfit.Space("obs1", -1, 1, label="$obs_1$ [GeV$^2$]")
    gauss = zfit.pdf.Gauss(mu=0.2, sigma=0.17, obs=obs)

    plt.figure()
    plt.title("Simple 1D plot using plot_model_pdf")
    plotter.plot_model_pdfV1(gauss, obs=obs)
    pytest.zfit_savefig(folder=folder)



def test_plot_sum_simple():
    obs = zfit.Space("obs1", -1, 1)
    gauss1 = zfit.pdf.Gauss(mu=0.2, sigma=0.17, obs=obs, label="Gauss1")
    gauss2 = zfit.pdf.Gauss(mu=-0.2, sigma=0.17, obs=obs, label="Gauss2")
    sum_pdf = zfit.pdf.SumPDF([gauss1, gauss2], fracs=0.5, label="Model")

    plt.figure()
    plt.title("Simple sum plot using plot_sumpdf_components_pdfV1")
    sum_pdf.plot.plotpdf()
    sum_pdf.plot.comp.plotpdf()
    pytest.zfit_savefig(folder=folder)



def test_plot_3d_simple():
    obs1 = zfit.Space("obs1", 10, 100)
    obs2 = zfit.Space("obs2", -5, 4)
    obs3 = zfit.Space("obs3", 0, 1)
    gauss1 = zfit.pdf.Gauss(mu=49, sigma=4, obs=obs1)
    gauss2 = zfit.pdf.Gauss(mu=-3, sigma=0.9, obs=obs2)
    gauss3 = zfit.pdf.Gauss(mu=0.2, sigma=0.07, obs=obs3)
    model = zfit.pdf.ProductPDF([gauss1, gauss2, gauss3])

    with pytest.raises(ValueError):
        plotter.plot_model_pdfV1(model)
    obsplot = zfit.Space("obs2", -5, -3)
    plt.figure()
    plt.title("Simple 3D plot using plot_model_pdf with space")
    plotter.plot_model_pdfV1(model, obs=obsplot)
    pytest.zfit_savefig(folder=folder)

    plt.figure()
    plt.title("Simple 3D plot using plot_model_pdf")
    plotter.plot_model_pdfV1(model, obs="obs2")
    pytest.zfit_savefig(folder=folder)

    model_ext = model.create_extended(1000)
    plt.figure()
    plt.title("Extended model 3D")
    with pytest.raises(ValueError):
        plotter.plot_model_pdfV1(model_ext)
    plotter.plot_model_pdfV1(model_ext, obs="obs2", extended=True)
    pytest.zfit_savefig(folder=folder)



#  Copyright (c) 2024 zfit
from __future__ import annotations

import functools as _functools
import io
from typing import Optional

import dill as __dill

# noinspection PyUnresolvedReferences
from dill import *  # noqa: F403

_NONE = object()


class ZfitDillDumpError(Exception):
    pass


class ZfitDillLoadError(Exception):
    pass


def __retry_with_gc(func, kwargs, *, max_retries: int | None = None):
    """Helper function to retry a function with garbage collection if necessary.

    Main intended use is to retry a function that failed due to a graph that cannot be pickled. This can happen
    if the graph is not garbage collected and still in memory. This function will retry the function after
    garbage collection. For the dill.dump and dill.dumps functions. Also catches the ipyopt error and raises a
    more informative error message.

    Args:
        func: Function to retry
        kwargs: Keyword arguments to pass to the function
        max_retries: Maximum number of retries. If None, defaults to 2.
    """
    redo = False
    if max_retries is None:
        max_retries = 2
    max_retries_reached = False
    result_dilled = _NONE
    funcgraph_phrase = "cannot pickle 'FuncGraph' object"
    dictchange_phrase = "dictionary changed size during iteration"
    ipyopt_phrase = "cannot pickle 'ipyopt.Problem' object"
    weakgone_phrase = "weak object has gone away"
    ipyopt_error_msg = (
        f"Cannot pickle {kwargs.get('obj')}, as it contains the ipyopt minimizer, which cannot be pickled/dilled."
        f"This is a known limitation. Use another method to dump the object: if you're dumping a FitResult, call `result.freeze()`"
        f" on it first, or use `zfit.hs3` for human readable serialization."
    )
    try:
        result_dilled = func(**kwargs)
    except TypeError as error:
        if not (redo := funcgraph_phrase in str(error) or weakgone_phrase in str(error)) or (
            max_retries_reached := max_retries <= 0
        ):
            if ipyopt_phrase in str(error):
                from zfit.exception import IpyoptPicklingError

                raise IpyoptPicklingError(ipyopt_error_msg) from error
            if max_retries_reached:
                msg = (
                    f"Max retries reached when dumping {kwargs}, error still occurred. If this happens more than once, please report on github in an issue."
                    f" The following error is not the first one to occur, to debug, please rerun with `max_retries=0` and increase the number`."
                )
                raise RuntimeError(msg) from error
            raise error
    except RuntimeError as error:
        if dictchange_phrase in str(error):
            redo = True
        else:
            raise error
    # need to get out, otherwise "error" has a reference to graph and gc can't collect it
    if redo:
        import gc

        gc.collect()
        result_dilled = __retry_with_gc(func, kwargs, max_retries=max_retries - 1)

    assert result_dilled is not _NONE, "result_dilled should have been set, logic error in zfit/dill.py"
    return result_dilled


@_functools.wraps(__dill.dumps)
def dumps(
    obj,
    protocol=None,
    byref=False,
    fmode=None,
    recurse=None,
    *,
    max_retries: Optional[int | bool] = None,
    verify: Optional[bool] = None,
    **kwds,
):
    """Wrapper around :py:func`dill.dumps` that helps dumping zfit objects as it retries with garbage collection if
    necessary.

    .. note::

        This function can dump any python object, including list/dicts of zfit objects. If two objects are in any way
        related, for example they share parameters, they *have* to be dumped together. Otherwise, the parameters will
        be duplicated when loading, with the same name, which will fail to run the fit.

    Additional argument max_retries: Maximum number of retries if it fails (can occur due to garbage collector required to run first).
        If None, defaults to 2.

    Original docstring:
    {docstring}
    """
    redo = False
    if max_retries is None:
        max_retries = 2
    elif max_retries < 0:
        msg = "max_retries has to be >= 0"
        raise ValueError(msg)
    if verify is None:
        verify = True
    kwargs = dict(obj=obj, protocol=protocol, byref=byref, fmode=fmode, recurse=recurse, **kwds)
    for ntry in range(max_retries + 1):
        out = __retry_with_gc(func=__dill.dumps, kwargs=kwargs, max_retries=max_retries)
        if verify:
            try:
                loads(out, max_retries=max_retries)
            except ZfitDillLoadError as error:
                if ntry == max_retries:
                    msg = "Tried to verify dumps by loading but failed."
                    raise ZfitDillDumpError(msg) from error
                redo = True
                from zfit import run

                run.clear_graph_cache(call_gc=True)
        if not redo:
            break

    return out


dumps.__doc__ = dumps.__doc__.format(docstring=__dill.dumps.__doc__)


@_functools.wraps(__dill.dump)
def dump(
    obj,
    file,
    protocol=None,
    byref=None,
    fmode=None,
    recurse=None,
    *,
    max_retries: Optional[int | bool] = None,
    verify: Optional[bool] = None,
    **kwds,
):
    """Wrapper around :py:func`dill.dump` that helps dumping zfit objects as it retries with garbage collection if
    necessary.

    .. note::

        This function can dump any python object, including list/dicts of zfit objects. If two objects are in any way
        related, for example they share parameters, they *have* to be dumped together. Otherwise, the parameters will
        be duplicated when loading, with the same name, which will fail to run the fit.

    Additional arguments
        - max_retries: Maximum number of retries if it fails (can occur due to garbage collector required to run first).
           If None, defaults to 2.
        - verify: **requires `file` to be in write *and* read mode (i.e. `w+b` instead of `wb` only).**
            This *can* clear the graph cache and should not have an effect on the results but may significantly slow down
            subsequent calls to the same zfit fits, as the graph needs to be recompiled.
            If True, the dump will be verified by loading it again. If it fails, the dump will be redone.
            If it fails after the maximum number of retries, a ZfitDillDumpError will be raised.

    Original docstring:
    {docstring}
    """
    redo = False
    if max_retries is None:
        max_retries = 2
    elif max_retries < 0:
        msg = "max_retries has to be >= 0"
        raise ValueError(msg)
    if verify is None:
        verify = True
    kwargs = dict(obj=obj, file=file, protocol=protocol, byref=byref, fmode=fmode, recurse=recurse, **kwds)
    initial_position = file.tell()
    for ntry in range(max_retries + 1):
        out = __retry_with_gc(func=__dill.dump, kwargs=kwargs, max_retries=max_retries)
        if verify:
            post_out_position = file.tell()
            file.seek(initial_position)
            try:
                load(file=file, max_retries=max_retries)
            except ZfitDillLoadError as error:
                if ntry == max_retries:
                    msg = "Tried to verify dumps by loading but failed."
                    raise ZfitDillDumpError(msg) from error
                redo = True
                from zfit import run

                run.clear_graph_cache(call_gc=True)
                file.seek(initial_position)  # reset to initial position
            else:
                redo = False
                file.seek(post_out_position)  # we're done verifying
        if not redo:
            break
    return out


def __retry_with_graphclear(func, kwargs, max_retries, _file_to_reset=None):
    original_error = None

    # make sure to reset the file to the initial position after an error was raised
    # and the read is re-attempted
    initial_position = None
    if _file_to_reset is not None:
        initial_position = _file_to_reset.tell()
    for i in range(max_retries + 1):
        if _file_to_reset is not None:
            _file_to_reset.seek(initial_position)
        try:
            out = func(**kwargs)
        except io.UnsupportedOperation as error:
            if "read" in str(error):
                msg = (
                    "Tried to verify (use `verify=False` to not verify) dumping by loading but failed (most likely because the file was opened in write mode only. "
                    "Try to open it in write and read mode, for example by changing `wb` to `w+b`.)"
                )
                raise io.UnsupportedOperation(msg) from error
        except Exception as error:
            if original_error is None:
                original_error = error
            if i == max_retries:
                msg = (
                    f"Max retries reached when loading {kwargs}, error still occurred. Original error {original_error}"
                )
                raise ZfitDillLoadError(msg) from error
            from zfit import run

            run.clear_graph_cache(call_gc=True)
        else:
            break
    return out


@_functools.wraps(__dill.loads)
def loads(str, *, max_retries: Optional[int | bool] = None, **kwds):
    """Wrapper around :py:func`dill.loads`that helps loading zfit objects as it retries with graph clearing if
    necessary.

    .. warning ::

        This function *may* clears the cached graph/traced functions. This should not have an effect on the
        results but may significantly slow down subsequent calls to the same zfit fits, as the graph needs to be
        recompiled.

    Additional argument max_retries: Maximum number of retries if it fails (can occur due to garbage collector required to run first).
        If None, defaults to 2.

    Original docstring:
    {docstring}
    """
    if max_retries is None:
        max_retries = 2
    elif max_retries < 0:
        msg = "max_retries has to be >= 0"
        raise ValueError(msg)
    kwargs = dict(str=str, **kwds)
    return __retry_with_graphclear(func=__dill.loads, kwargs=kwargs, max_retries=max_retries)


@_functools.wraps(__dill.load)
def load(file, *, max_retries: Optional[int | bool] = None, **kwds):
    """Wrapper around :py:func`dill.load`that helps loading zfit objects as it retries with graph clearing if necessary.

    .. warning ::

        This function *may* clears the cached graph/traced functions. This should not have an effect on the
        results but may significantly slow down subsequent calls to the same zfit fits, as the graph needs to be
        recompiled.

    Additional argument max_retries: Maximum number of retries if it fails (can occur due to garbage collector required to run first).
        If None, defaults to 2.

    Original docstring:
    {docstring}
    """
    if max_retries is None:
        max_retries = 2
    elif max_retries < 0:
        msg = "max_retries has to be >= 0"
        raise ValueError(msg)
    kwargs = dict(file=file, **kwds)
    return __retry_with_graphclear(func=__dill.load, kwargs=kwargs, max_retries=max_retries, _file_to_reset=file)


dump.__doc__ = dump.__doc__.format(docstring=__dill.dump.__doc__)
